import numpy as np
from scipy.ndimage import affine_transform
try: from numba import njit as jit
except:
    print('install numba may be several times faster!')
    jit = None
# jit = None
    
def affine_jit(img, m, offset, output_shape=0, output=0, order=0, prefilter=0):
    kr=m[0]; kc=m[1]; ofr=offset[0]; ofc=offset[1];
    for r in range(output_shape[0]):
        for c in range(output_shape[1]):
            rr = int(r*kr+ofr)
            cc = int(c*kc+ofc)
            output[r,c] = img[rr,cc]
if not jit is None: affine_transform = jit(affine_jit)

def blend(img, out, msk, mode):
    if mode=='set': out[:] = img
    if mode=='min': np.minimum(out, img, out=out)
    if mode=='max': np.maximum(out, img, out=out)
    if mode=='msk':
        msk = np.logical_not(msk)
        out.T[:] *= msk.T
        out += img
    if isinstance(mode, float):
        np.multiply(out, 1-mode, out=out, casting='unsafe')
        np.multiply(img, mode, out=img, casting='unsafe')
        out += img

if not jit is None:
    @jit
    def blend_set(img, out, msk, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                out[r,c] = img[r,c]
    @jit
    def blend_min(img, out, msk, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                out[r,c] = min(img[r,c], out[r,c])
    @jit
    def blend_max(img, out, msk, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                out[r,c] = max(img[r,c], out[r,c])
    @jit
    def blend_msk(img, out, msk, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                if img[r,c] != 0: out[r,c] = img[r,c]
    @jit
    def blend_mix(img, out, msk, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                out[r,c] = img[r,c]*mode + out[r,c]*(1-mode)

    def blend_jit(img, out, msk, mode):
        if mode=='set': blend_set(img, out, msk, mode)
        if mode=='min': blend_min(img, out, msk, mode)
        if mode=='max': blend_max(img, out, msk, mode)
        if mode=='msk': blend_msk(img, out, msk, mode)
        if isinstance(mode, float): blend_mix(img, out, msk, mode)
    blend = blend_jit

def stretch(img, out, rg, log=False):
    if img.dtype==np.uint8 and not log and (rg==[(0,255)] or rg==(0,255)):
        out[:] = img
    elif not log:
        ptp = max(rg[1]-rg[0], 1e-6)
        np.clip(img, rg[0], rg[1], out=img)
        np.subtract(img, rg[0], out=img, casting='unsafe')
        np.multiply(img, 255/ptp, out=out, casting='unsafe')
    elif img.itemsize<3:
        length = 2**(img.itemsize*8)
        lut = np.arange(length, dtype=np.float32)
        if img.dtype in (np.int8, np.int16):
            lut[length//2:] -= length
        np.clip(lut, rg[0], rg[1], out=lut)
        np.subtract(lut, rg[0]-1, out=lut)
        ptp = np.log(max(rg[1]-rg[0]+1, 1+1e-6))
        np.log(lut, out=lut)
        lut *= 255/np.log(max(rg[1]-rg[0]+1, 1+1e-6))
        out[:] = lut[img]
    else:
        fimg = img.ravel().view(np.float32)
        fimg = fimg[:img.size].reshape(img.shape)
        np.clip(img, rg[0], rg[1], out=fimg)
        np.subtract(fimg, rg[0]-1, out=fimg)
        ptp = np.log(max(rg[1]-rg[0]+1, 1+1e-6))
        np.log(fimg, out=fimg)
        np.multiply(fimg, 255/ptp, out=out, casting='unsafe')

if not jit is None:
    @jit
    def stretch_linear(img, out, rg):
        ptp = max(rg[1]-rg[0], 1e-6)
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                v = (img[r,c]-rg[0])/ptp*255
                out[r,c] = min(max(v, 0), 255)
    @jit
    def stretch_log(img, out, rg):
        ptp = 255/np.log(max(rg[1]-rg[0]+1, 1+1e-6))
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                v = np.log(img[r,c]-rg[0]+1)*ptp
                out[r,c] = min(max(v, 0), 255)
    @jit
    def stretch_lut(img, out, lut):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                out[r,c] = lut[img[r,c]]
        
    def stretch_jit(img, out, rg, log=False):
        if img.dtype==np.uint8 and not log and (rg==[(0,255)] or rg==(0,255)):
            out[:] = img
        elif not log:
            stretch_linear(img, out, rg)
        elif img.itemsize<3:
            length = 2**(img.itemsize*8)
            lut = np.arange(length, dtype=np.float32)
            if img.dtype in (np.int8, np.int16):
                lut[length//2:] -= length
            np.clip(lut, rg[0], rg[1], out=lut)
            np.subtract(lut, rg[0]-1, out=lut)
            ptp = np.log(max(rg[1]-rg[0]+1, 1+1e-6))
            np.log(lut, out=lut)
            lut *= 255/np.log(max(rg[1]-rg[0]+1, 1+1e-6))
            stretch_lut(img, out, lut)
        else:
            stretch_log(img, out, rg)
            
    stretch = stretch_jit

def complex_norm(ori, real, img, out):
    np.abs(ori, out=out)
    return out

if not jit is None:
    @jit
    def complex_norm(ori, real, img, out):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                out[r,c] = (img[r,c]**2+real[r,c]**2)**0.5
        return out

def lookup(img, lut, out, mode='set'):
    blend(lut[img], out, img, mode)

if not jit is None:
    @jit
    def lookup_set(img, lut, out, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                for i in (0,1,2):
                    out[r,c,i] = lut[img[r,c],i]
    @jit
    def lookup_min(img, lut, out, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                for i in (0,1,2):
                    out[r,c,i] = min(lut[img[r,c],i], out[r,c,i])
    @jit
    def lookup_max(img, lut, out, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                for i in (0,1,2):
                    out[r,c,i] = max(lut[img[r,c],i], out[r,c,i])
    @jit
    def lookup_msk(img, lut, out, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                if img[r,c] != 0:
                    for i in (0,1,2): out[r,c,i] = 0
                for i in (0,1,2):
                    out[r,c,i] += lut[img[r,c],i]
    @jit
    def lookup_mix(img, lut, out, mode):
        for r in range(img.shape[0]):
            for c in range(img.shape[1]):
                for i in (0,1,2):
                    out[r,c,i] = lut[img[r,c],i]*mode + out[r,c,i]*(1-mode)

    def lookup_jit(img, lut, out, mode):
        if mode == 'set': lookup_set(img, lut, out, mode)
        if mode == 'msk': lookup_msk(img, lut, out, mode)
        if mode == 'max': lookup_max(img, lut, out, mode)
        if mode == 'min': lookup_min(img, lut, out, mode)
        if isinstance(mode, float): lookup_mix(img, lut, out, mode)
        
    lookup = lookup_jit

# mode: set, min, max, mix, nor
def mix_img(img, m, o, shp, buf, rgb, byt, rg=(0,255), lut=None, log=True, cns=0, mode='set'):
    if img is None: return
    img = img.reshape((img.shape[0], img.shape[1], -1))
    if isinstance(rg, tuple): rg = [rg]*img.shape[2]

    if isinstance(cns, int):
        if np.iscomplexobj(buf):
            affine_transform(img[:,:,0].real, m, o, shp, buf.real, 0, prefilter=False)
            affine_transform(img[:,:,0].imag, m, o, shp, buf.imag, 0, prefilter=False)
            buf = complex_norm(buf, buf.real, buf.imag, buf.real)
        else: 
            affine_transform(img[:,:,cns], m, o, shp, buf, 0, prefilter=False)
        stretch(buf, byt, rg[cns], log)
        return lookup(byt, lut, rgb, mode)
    for i,v in enumerate(cns):
        if v==-1: rgb[:,:,i] = 0
        elif mode=='set' and img.dtype==np.uint8 and rg[v]==(0,255) and not log:
            affine_transform(img[:,:,v], m, o, shp, rgb[:,:,i], 0, prefilter=False)
        else:
            affine_transform(img[:,:,v], m, o, shp, buf, 0, prefilter=False)
            stretch(buf, byt, rg[v], log)
            blend(byt, rgb[:,:,i], byt, mode)

'''
rgb = np.array([[[0,0,0]]], dtype=np.uint8)
lut = np.array([[0,0,0]], dtype=np.uint8)
mix_img(img, (1,1), (0,0), (1,1), img, rgb, img, (0,255), lut, 0, 'set')
'''



import numpy as np
from skimage import draw
from ..object.shape import *

def offset(shp, dx, dy):
	if shp.dtype in {'rectangle', 'ellipse', 'circle'}:
		shp.body[:2] += dx, dy
	elif shp.dtype in {'rectangles', 'ellipses', 'circles'}:
		shp.body[:,:2] += dx, dy
	elif isinstance(shp, np.ndarray):
		shp += dx, dy
	elif isinstance(shp.body, list):
		for i in shp.body: offset(i, dx, dy)

def mark2shp(mark):
    style = mark.copy()
    style.pop('body')
    keys = {'point':Point, 'points':Points, 'line':Line, 'lines':Lines,
            'polygon':Polygon, 'polygons':Polygons, 'circle':Circle,
            'circles':Circles, 'rectangle':Rectangle, 'rectangles':Rectangles,
            'ellipse':Ellipse, 'ellipses':Ellipses, 'text':Text, 'texts':Texts}
    if mark['type'] in keys: return keys[mark['type']](mark['body'], **style)
    if mark['type']=='layer':
        return Layer([mark2shp(i) for i in mark['body']], **style)
    if mark['type']=='layers':
        return Layers(dict(zip(mark['body'].keys(),
            [mark2shp(i) for i in mark['body'].values()])), **style)

def json2shp(obj):
    if obj['type']=='Point':
        return Point(obj['coordinates'])
    if obj['type']=='MultiPoint':
        return Points(obj['coordinates'])
    if obj['type']=='LineString':
        return Line(obj['coordinates'])
    if obj['type']=='MultiLineString':
        return Lines(obj['coordinates'])
    if obj['type']=='Polygon':
        return Polygon(obj['coordinates'])
    if obj['type']=='MultiPolygon':
        return Polygons(obj['coordinates'])
    if obj['type']=='GeometryCollection':
        return Layer([json2shp(i) for i in obj['geometries']])

def geom2shp(obj): return json2shp(geom.mapping(obj))

def geom_flatten(obj, geoms=None):
    geoms, root = ([], True) if geoms is None else (geoms, False)
    if isinstance(obj, geom.GeometryCollection):
        for i in obj: geom_flatten(i, geoms)
    elif type(obj) in {geom.MultiPolygon, geom.MultiPoint, geom.MultiLineString}: 
        geoms.extend(list(obj))
    else: geoms.append(obj)
    if root: return geom.GeometryCollection(geoms)

def geom_union(obj):
    return geom_flatten(unary_union(geom_flatten(obj)))


def draw_circle(r):
    xs, ys = np.mgrid[-r:r+1, -r:r+1]
    rcs = np.where((xs**2 + ys**2)<r+0.5)
    return np.array(rcs)-r

def draw_thick(rr, cc, r, shp):
    rs, cs = draw_circle(r)
    rr, cc = rr[:,None] + rs, cc[:,None] + cs
    msk = (rr>=0) & (rr<shp[0]) & (cc>=0) & (cc<shp[1])
    return rr[msk], cc[msk]

def draw_line(r1, c1, r2, c2, shp, lw=0):
    rr, cc = draw.line(r1, c1, r2, c2)
    return draw_thick(rr, cc, lw, shp)

def draw_lines(rs, cs, shp, lw):
    rr, cc = [], []
    for r1, c1, r2, c2 in zip(rs[:-1], cs[:-1], rs[1:], cs[1:]):
        r, c = draw.line(r1, c1, r2, c2)
        rr.append(r); cc.append(c)
    rr, cc = np.hstack(rr), np.hstack(cc)
    return draw_thick(rr, cc, lw, shp)

def draw_path(rs, cs, shp, lw):
    rr, cc = draw.polygon_perimeter(rs, cs, shp, True)
    return draw_thick(rr, cc, lw, shp)

def draw_shp(shp, img=None, color=1, lw=1):
    if isinstance(img, tuple):
        img = np.zeros(img, dtype=np.int8)
    if type(shp) in {geom.Point, geom.MultiPoint}:
        arr = np.array(shp).reshape((-1,2)).T
        cc, rr = np.round(arr).astype(np.int32)
        rr, cc = draw_thick(rr, cc, lw, img.shape)
        img[rr, cc] += color
    if type(shp) is geom.LineString:
        cc, rr = np.round(np.array(shp).T).astype(np.int32)
        rr, cc = draw_lines(rr, cc, img.shape, lw)
        img[rr, cc] += color
    if type(shp) is geom.LinearRing:
        cc, rr = np.round(np.array(shp).T).astype(np.int32)
        if lw>0: rr, cc = draw_lines(rr, cc, img.shape, lw)
        else: rr, cc = draw.polygon(rr, cc, img.shape)
        img[rr, cc] += color
    if type(shp) is geom.MultiLineString:
        for i in shp: draw_shp(i, img, color, lw)
    if type(shp) is geom.Polygon and lw>0:
        draw_shp(shp.exterior, img, color, lw)
        for i in shp.interiors: draw_shp(i, img, color, lw)
    if type(shp) is geom.Polygon and lw==0:
        draw_shp(shp.exterior, img, color, lw)
        for i in shp.interiors: draw_shp(i, img, -color, lw)
    if type(shp) is geom.MultiPolygon:
        for i in shp: draw_shp(i, img, color, lw)
    if type(shp) is geom.GeometryCollection:
        for i in shp: draw_shp(i, img, color, lw)
    return img


from .shp_tools import *
from .shpbase import BaseEditor
from ..tolact import ImageTool
from ...object import ROI

class BaseROI(ImageTool):
	def __init__(self, base): 
		base.__init__(self)
		self.base = base

	def mouse_down(self, img, x, y, btn, **key):
		if img.roi is None: img.roi = ROI()
		else: img.roi.msk = None
		self.base.mouse_down(self, img.roi, x, y, btn, **key)

	def mouse_up(self, img, x, y, btn, **key):
		self.base.mouse_up(self, img.roi, x, y, btn, **key)
		if not img.roi is None:
			if len(img.roi.body)==0: img.roi = None
			else: img.roi.msk = None

	def mouse_move(self, img, x, y, btn, **key):
		self.base.mouse_move(self, img.roi, x, y, btn, **key)

	def mouse_wheel(self, img, x, y, d, **key):
		self.base.mouse_wheel(self, img.roi, x, y, d, **key)

class PolygonROI(BaseROI):
	title = 'Polygon ROI'
	def __init__(self): 
		BaseROI.__init__(self, PolygonEditor)

class LineROI(BaseROI):
	title = 'Line ROI'
	def __init__(self): 
		BaseROI.__init__(self, LineEditor)

class PointROI(BaseROI):
	title = 'Point ROI'
	def __init__(self): 
		BaseROI.__init__(self, PointEditor)

class RectangleROI(BaseROI):
	title = 'Rectangle ROI'
	def __init__(self): 
		BaseROI.__init__(self, RectangleEditor)

class EllipseROI(BaseROI):
	title = 'Ellipse ROI'
	def __init__(self): 
		BaseROI.__init__(self, EllipseEditor)

class FreeLineROI(BaseROI):
	title = 'Free Line ROI'
	def __init__(self): 
		BaseROI.__init__(self, FreeLineEditor)

class FreePolygonROI(BaseROI):
	title = 'Free Polygon ROI'
	def __init__(self): 
		BaseROI.__init__(self, FreePolygonEditor)


# -*- coding: utf-8 -*-
"""
Created on Sat Dec  3 03:32:05 2016
@author: yxl
"""
import threading
from time import time

class Table:
    title = 'TableFilter'
    note = []
    para = None
    'req_sel, req_row, req_col, auto_snap, auto_msk, msk_not, num_only, preview'
    view = None
    prgs = None
    modal = True
    asyn = True

    def __init__(self, tps=None):
        
        self.dialog = None
    
    def progress(self, i, n): self.prgs = int(i*100/n)

    def load(self, tps):return True
        
    def preview(self, tps, para):
        self.run(tps, tps.snap, tps.data, para)
        tps.update()
    
    def show(self):
        preview = lambda para, tps=self.tps: self.preview(tps, para) or tps.update()
        return self.app.show_para(self.title, self.para, self.view, preview, 
            on_ok=lambda : self.ok(self.tps), on_help=self.on_help,
            on_cancel=lambda : self.cancel(self.tps) or self.tps.update(), 
            preview='preview' in self.note, modal=self.modal)
    
    def run(self, tps, snap, data, para = None):pass
        
    def cancel(self, tps):
        if 'auto_snap' in self.note:
            tps.data[tps.snap.columns] = tps.snap
            tps.update()

    def on_help(self):
        self.app.show_md(self.__doc__ or 'No Document!', self.title)

    def ok(self, tps, para=None, callafter=None):
        if para == None: para = self.para
        if self.asyn and self.app.asyn:
            threading.Thread(target = self.runasyn, 
                args = (tps, tps.data, tps.snap, para, callafter)).start()
        else: self.runasyn(tps, tps.data, tps.snap, para, callafter)

    def runasyn(self,  tps, snap, data, para = None, callback = None):
        self.app.record_macros('{}>{}'.format(self.title, para))
        self.app.add_task(self)
        start = time()
        self.run(tps, data, snap, para)
        self.app.info('%s: cost %.3fs'%(tps.title, time()-start))
        tps.update()
        self.app.remove_task(self)
        if callback!=None:callback()

    def check(self, tps):
        print(self.note)
        if tps == None:
            self.app.alert('no table opened!')
            return False
        if 'req_sel' in self.note:
            print(tps.rowmsk, tps.colmsk)
            if isinstance(tps.rowmsk, slice) and\
             isinstance(tps.colmsk, slice):
                self.app.alert('no selection!')
                return False
        if 'req_row' in self.note:
            print(tps.rowmsk, tps.colmsk)
            if isinstance(tps.rowmsk, slice):
                self.app.alert('need row selection!')
                return False
        if 'req_col' in self.note:
            print(tps.rowmsk, tps.colmsk)
            if isinstance(tps.colmsk, slice):
                self.app.alert('need col selection!')
                return False
        return True

    def start(self, app, para=None, callback=None):
        self.app, self.tps = app, app.get_table()
        #print self.title, para
        if not self.check(self.tps):return
        if not self.load(self.tps):return
        if 'auto_snap' in self.note:
            if 'auto_msk' in self.note: mode = True
            elif 'msk_not' in self.note: mode = False
            else: mode = None
            self.tps.snapshot(mode, 'num_only' in self.note)
        if para!=None:
            self.ok(self.tps, para, callback)
        elif self.view==None:
            if not self.__class__.show is Table.show:
                if self.show():
                    self.ok(self.tps, para, callback)
            else: self.ok(self.tps, para, callback)
        elif self.modal:
            if self.show():
                self.ok(self.tps, para, callback)
            else:self.cancel(self.tps)
            if not self.dialog is None: self.dialog.Destroy()
        else: self.show()


import tkinter as tk
import time, threading
import sys
sys.path.append('/media/hzy/程序/novalide/forgitcommit/scitk/scitk/')
from tkinter import ttk
from scitk.widgets import MenuBar, ToolBar, ParaDialog
from scitk.canvas import CanvasNoteBook
#from scitk.widgets import ProgressBar
from scitk.grid import GridFrame
#from scitk.mesh import Canvas3DFrame
#from scitk.text import MDFrame, TextFrame
from scitk.plot import PlotFrame
from sciapp import App, Source
from sciapp.object import Image, Table

class ImageApp(ttk.Frame, App):
    def __init__( self, parent ):
        ttk.Frame.__init__ (self,parent)
        #title = 'ImageApp', 
        #size = wx.Size(800,600), pos = wx.DefaultPosition, 
        # style = wx.RESIZE_BORDER|wx.DEFAULT_FRAME_STYLE|wx.TAB_TRAVERSAL )
        App.__init__(self)
##        self.SetSizeHints( wx.Size(600,-1) )

        sizer = ttk.Frame(self)#wx.BoxSizer(wx.VERTICAL)
        sizer.pack(expand=True,fill=tk.BOTH)
##        self.toolbar = ToolBar(self)
##        self.toolbar.Fit()
##        sizer.Add(self.toolbar, 0, wx.EXPAND |wx.ALL, 0)

        self.canvasnb = CanvasNoteBook(sizer)
##        self.canvasnb.Bind( wx.lib.agw.aui.EVT_AUINOTEBOOK_PAGE_CHANGED, self.on_active_img)
##        self.canvasnb.Bind( wx.lib.agw.aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.on_close_img)
        self.canvasnb.pack(fill=tk.BOTH,expand=True)
        #sizer.Add(self.canvasnb, 1, wx.EXPAND |wx.ALL, 0)

        sizersta = ttk.Frame(self)#wx.BoxSizer( wx.HORIZONTAL )
        sizersta.pack(expand=True,fill=tk.BOTH)
        self.stapanel = stapanel = ttk.Frame(self)
        
        self.txt_info = ttk.Label( stapanel,text="ImageApp")
        self.txt_info.pack(expand=True,fill=tk.BOTH)
       # sizersta.Add( self.txt_info, 1, wx.ALIGN_BOTTOM|wx.BOTTOM|wx.LEFT|wx.RIGHT, 2 )
        self.pro_bar = ProgressBar(stapanel)
        sizersta.Add( self.pro_bar, 0, wx.ALL, 2 )
        stapanel.SetSizer(sizersta)
        sizer.Add(self.stapanel, 0, wx.EXPAND | wx.ALL, 0)
        self.SetSizer(sizer)
        self.Layout()
        self.Centre( wx.BOTH )
        self.Bind(wx.EVT_CLOSE, self.on_close)
        
    def record_macros(self, x): print(x)
    
    def add_task(self, task):
        self.task_manager.add(task.title, task)
        tasks = self.task_manager.gets()
        tasks = [(p.title, lambda t=p:p.prgs) for n,p,t in tasks]
        self.pro_bar.SetValue(tasks)

    def remove_task(self, task):
        self.task_manager.remove(obj=task)
        tasks = self.task_manager.gets()
        tasks = [(p.title, lambda t=p:p.prgs) for n,p,t in tasks]
        self.pro_bar.SetValue(tasks)

    def on_active_img(self, event):
        self.active_img(self.canvasnb.canvas().image.name)
        # self.add_img_win(self.canvasnb.canvas())

    def on_close_img(self, event):
        canvas = event.GetEventObject().GetPage(event.GetSelection())
        App.close_img(self, canvas.image.title)

    def on_new_tab(self, event):
        self.add_tab(event.GetEventObject().grid.table)
        self.add_tab_win(event.GetEventObject().grid)

    def on_close_tab(self, event):
        self.remove_tab_win(event.GetEventObject().grid)
        self.remove_tab(event.GetEventObject().grid.table)
        event.Skip()
        
    def info(self, value):
        wx.CallAfter(self.txt_info.SetLabel, value)

    def set_progress(self, value):
        v = max(min(value, 100), 0)
        self.pro_bar.SetValue(v)
        if value==-1:
            self.pro_bar.Hide()
        elif not self.pro_bar.IsShown():
            self.pro_bar.Show()
            self.stapanel.GetSizer().Layout()
        self.pro_bar.Update()

    def on_close(self, event):
        self.Destroy()
        sys.exit()

    def _show_img(self, img, title=None):
        canvas = self.canvasnb.add_canvas()
        if not isinstance(img, Image): 
            img = Image(img, title)
        App.show_img(self, img, img.title)
        canvas.set_img(img)

    def show_img(self, img, title=None):
        wx.CallAfter(self._show_img, img, title)

    def _show_table(self, tab, title):
        cframe = GridFrame(self)
        grid = cframe.grid
        grid.set_data(tab)
        if not title is None:
            grid.table.name = title
        cframe.Bind(wx.EVT_ACTIVATE, self.on_new_tab)
        cframe.Bind(wx.EVT_CLOSE, self.on_close_tab)
        cframe.Show()

    def show_table(self, tab, title=None):
        wx.CallAfter(self._show_table, tab, title)

    def show(self, tag, cont, title):
        tag = tag or 'img'
        if tag=='img':
            self.show_img([cont], title)
        elif tag=='imgs':
            self.show_img(cont, title)
        elif tag=='tab':
            self.show_table(cont, title)
        elif tag=='mc':
            self.run_macros(cont)
        elif tag=='md':
            self.show_md(cont, title)
        elif tag=='wf':
            self.show_workflow(cont, title)
        else: self.alert('no view for %s!'%tag)

    def info(self, cont): 
        wx.CallAfter(self.txt_info.SetLabel, cont)

    def _alert(self, info, title='ImagePy'):
        dialog=wx.MessageDialog(self, info, title, wx.OK)
        dialog.ShowModal() == wx.ID_OK
        dialog.Destroy()

    def alert(self, info, title='ImagePy'):
        wx.CallAfter(self._alert, info, title)

    def yes_no(self, info, title='ImagePy'):
        dialog = wx.MessageDialog(self, info, title, wx.YES_NO | wx.CANCEL)
        rst = dialog.ShowModal()
        dialog.Destroy()
        dic = {wx.ID_YES:'yes', wx.ID_NO:'no', wx.ID_CANCEL:'cancel'}
        return dic[rst]

    def get_path(self, title, filt, io, name=''):
        filt = '|'.join(['%s files (*.%s)|*.%s'%(i.upper(),i,i) for i in filt])
        dic = {'open':wx.FD_OPEN, 'save':wx.FD_SAVE}
        dialog = wx.FileDialog(self, title, '', name, filt, dic[io])
        rst = dialog.ShowModal()
        path = dialog.GetPath() if rst == wx.ID_OK else None
        dialog.Destroy()
        return path

    def show_para(self, title, para, view, on_handle=None, on_ok=None, on_cancel=None, preview=False, modal=True):
        dialog = ParaDialog(self, title)
        dialog.init_view(view, para, preview, modal=modal, app=self)
        dialog.Bind('cancel', on_cancel)
        dialog.Bind('parameter', on_handle)
        dialog.Bind('commit', on_ok)
        return dialog.show()

    def start(imgs=[], plgs=[]):
        app = tk.Tk()
        frame = ImageApp(None)
        for name, i in imgs: frame.show_img([i], name)
        for name, i in plgs: frame.toolbar.add_tool(name, i)

        app.mainloop()

if __name__ == '__main__':
    from skimage.data import camera
    from scitk.plugins.filters import Gaussian

    ImageApp.start(
        imgs = [('camera', camera())], 
        plgs=[('G', Gaussian)])


import threading
from time import time

class Free:
    title = 'Free'
    view = None
    para = None
    prgs = None
    asyn = True

    def progress(self, i, n): self.prgs = int(i*100/n)

    def run(self, para=None): print('this is a plugin')

    def runasyn(self, para, callback=None):
        self.app.add_task(self)
        self.app.record_macros('{}>{}'.format(self.title, para))
        start = time()
        self.run(para)
        self.app.info('%s: cost %.3fs'%(self.title, time()-start))
        self.app.remove_task(self)
        if callback!=None:callback()

    def load(self):return True

    def show(self):
        if self.view==None:return True
        return self.app.show_para(self.title, self.para, self.view, None)

    def start(self, app, para=None, callback=None):
        self.app = app
        if not self.load():return
        if para!=None or self.show():
            if para==None:para = self.para
            if self.asyn and app.asyn:
                threading.Thread(target = self.runasyn, 
                    args = (para, callback)).start()
            else: 
                self.runasyn(para, callback)


from .shpbase import *
from ...util import geom_flatten

def inbase(key, btn):
	status = key['ctrl'], key['alt'], key['shift']
	return status == (1,1,0) or btn in {2,3}

class PointEditor(BaseEditor):
	title = 'Point Tool'
	def __init__(self): 
		BaseEditor.__init__(self)

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn):
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if btn==1 and not key['alt'] and not key['ctrl']:
			shp.body.append(Point([x,y]))
			shp.dirty = True

class LineEditor(BaseEditor):
	title = 'Line Tool'
	def __init__(self): 
		BaseEditor.__init__(self)
		self.cur, self.n, self.obj = 0, 0, None

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn) and self.obj is None:
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1:
			if self.obj is None: 
				self.obj = Line([(x,y)])
				shp.body.append(self.obj)
			else: 
				self.obj.body = np.vstack((self.obj.body, [(x,y)]))
			anchor = Points(mark(self.obj)[0], color=(255,0,0))
			key['canvas'].marks['buffer'] = anchor
		if btn==3 and not self.obj is None :
			self.obj.body = np.vstack((self.obj.body, [(x,y)]))
			self.obj.dirty, shp.dirty, self.obj = True, True, None
			del key['canvas'].marks['buffer']
		shp.dirty = True

class PolygonEditor(BaseEditor):
	title = 'Polygon Tool'
	def __init__(self): 
		BaseEditor.__init__(self)
		self.cur, self.n, self.obj = 0, 0, None

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn) and self.obj is None:
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1:
			if self.obj is None: 
				self.obj = Line([(x,y)])
				shp.body.append(self.obj)
			else: 
				self.obj.body = np.vstack((self.obj.body, [(x,y)]))
			anchor = Points(mark(self.obj)[0], color=(255,0,0))
			key['canvas'].marks['buffer'] = anchor
		if btn==3 and not self.obj is None :
			body = np.vstack((self.obj.body, [(x,y)]))
			shp.body[-1] = Polygon(body)
			if key['alt'] or key['shift']:
				obj = shp.body.pop(-1)
				rst = geom_union(shp.to_geom())
				if key['alt'] and not key['shift']:
					rst = rst.difference(obj.to_geom())
				if key['shift'] and not key['alt']:
					rst = rst.union(obj.to_geom())
				if key['shift'] and key['alt']:
					rst = rst.intersection(obj.to_geom())
				layer = geom2shp(geom_flatten(rst))
				shp.body = layer.body
			self.obj, shp.dirty = None, True
			del key['canvas'].marks['buffer']
		shp.dirty = True

class RectangleEditor(BaseEditor):
	title = 'Rectangle Tool'
	def __init__(self): 
		BaseEditor.__init__(self)
		self.obj, self.p = None, None

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn):
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1:
			self.obj = Rectangle([x, y, 0, 0])
			self.p = x, y
			shp.body.append(self.obj)

	def mouse_up(self, shp, x, y, btn, **key): 
		if inbase(key, btn):
			BaseEditor.mouse_up(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if self.p == (x, y) and not self.obj is None:
			self.obj = self.p = None
			return shp.body.pop(-1)
		if (key['alt'] or key['shift']) and not self.obj is None:
			obj = shp.body.pop(-1)
			rst = geom_union(shp.to_geom())
			if key['alt'] and not key['shift']:
				rst = rst.difference(obj.to_geom())
			if key['shift'] and not key['alt']:
				rst = rst.union(obj.to_geom())
			if key['shift'] and key['alt']:
				rst = rst.intersection(obj.to_geom())
			layer = geom2shp(geom_flatten(rst))
			shp.body = layer.body
		self.obj, shp.dirty = None, True
		if 'buffer' in key['canvas'].marks:
			del key['canvas'].marks['buffer']

	def mouse_move(self, shp, x, y, btn, **key): 
		BaseEditor.mouse_move(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if not self.obj is None:
			self.obj.body[2:] = (x, y) - self.obj.body[:2]
			anchor = Points(mark(self.obj)[0], color=(255,0,0))
			key['canvas'].marks['buffer'] = anchor
			self.obj.dirty = shp.dirty = True

class EllipseEditor(BaseEditor):
	title = 'Ellipse Tool'
	def __init__(self): 
		BaseEditor.__init__(self)
		self.obj, self.p = None, None

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn):
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1:
			self.p = (x, y)
			self.obj = Ellipse([x, y, 0, 0, 0])
			shp.body.append(self.obj)

	def mouse_up(self, shp, x, y, btn, **key): 
		if inbase(key, btn):
			BaseEditor.mouse_up(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if self.p == (x, y) and not self.obj is None:
			self.obj = self.p = None
			return shp.body.pop(-1)
		if (key['alt'] or key['shift']) and not self.obj is None:
			obj = shp.body.pop(-1)
			rst = geom_union(shp.to_geom())
			if key['alt'] and not key['shift']:
				rst = rst.difference(obj.to_geom())
			if key['shift'] and not key['alt']:
				rst = rst.union(obj.to_geom())
			if key['shift'] and key['alt']:
				rst = rst.intersection(obj.to_geom())
			layer = geom2shp(geom_flatten(rst))
			shp.body = layer.body
		self.obj, shp.dirty = None, True
		if 'buffer' in key['canvas'].marks:
			del key['canvas'].marks['buffer']

	def mouse_move(self, shp, x, y, btn, **key): 
		BaseEditor.mouse_move(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if not self.obj is None:
			ox, oy = self.p
			self.obj.body[:] = [(x+ox)/2, (y+oy)/2, (x-ox)/2, (y-oy)/2, 0]
			anchor = Points(mark(self.obj)[0], color=(255,0,0))
			key['canvas'].marks['buffer'] = anchor
			shp.dirty = self.obj.dirty = True

class FreeLineEditor(BaseEditor):
	title = 'Free Line Tool'
	def __init__(self): 
		BaseEditor.__init__(self)
		self.cur, self.n, self.obj = 0, 0, None

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn):
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1 and self.obj is None: 
			self.obj = Line([(x,y)])
			shp.body.append(self.obj)
			shp.dirty = True

	def mouse_up(self, shp, x, y, btn, **key): 
		if inbase(key, btn):
			BaseEditor.mouse_up(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		self.obj = None

	def mouse_move(self, shp, x, y, btn, **key): 
		BaseEditor.mouse_move(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if not self.obj is None:
			self.obj.body = np.vstack((self.obj.body, [(x,y)]))
			self.obj.dirty = shp.dirty = True

class FreePolygonEditor(BaseEditor):
	title = 'Free Polygon Tool'
	def __init__(self): 
		BaseEditor.__init__(self)
		self.cur, self.n, self.obj = 0, 0, None

	def mouse_down(self, shp, x, y, btn, **key):
		if inbase(key, btn):
			BaseEditor.mouse_down(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1 and self.obj is None: 
			self.obj = Line([(x,y)])
			shp.body.append(self.obj)
			shp.dirty = True

	def mouse_up(self, shp, x, y, btn, **key): 
		if inbase(key, btn):
			BaseEditor.mouse_up(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if btn==1 and not self.obj is None:
			body = np.vstack((self.obj.body, self.obj.body[0]))
			shp.body[-1] = Polygon(body)
			body = np.vstack((self.obj.body, [(x,y)]))
			shp.body[-1] = Polygon(body)
			if key['alt'] or key['shift']:
				obj = shp.body.pop(-1)
				rst = geom_union(shp.to_geom())
				if key['alt'] and not key['shift']:
					rst = rst.difference(obj.to_geom())
				if key['shift'] and not key['alt']:
					rst = rst.union(obj.to_geom())
				if key['shift'] and key['alt']:
					rst = rst.intersection(obj.to_geom())
				layer = geom2shp(geom_flatten(rst))
				shp.body = layer.body
			self.obj, shp.dirty = None, True

	def mouse_move(self, shp, x, y, btn, **key): 
		BaseEditor.mouse_move(self, shp, x, y, btn, **key)
		if key['alt'] and key['ctrl']: return
		if not self.obj is None:
			self.obj.body = np.vstack((self.obj.body, [(x,y)]))
			self.obj.dirty = shp.dirty = True


import os
from sciapp import Source
from imagepy import root_dir
from . import Free, Simple, Table
#from .macros import Macros
from ... import Manager
import numpy as np

ReaderManager, WriterManager = Manager(), Manager()

class Reader(Free):
    para = {'path':''}
    tag = None

    def show(self):
        filt = [i.lower() for i in self.filt]
        self.para['path'] = self.app.get_path('Open..', filt, 'open', '')
        return not self.para['path'] is None

    #process
    def run(self, para = None):
        #add_recent(para['path'])
        fp, fn = os.path.split(para['path'])
        fn, fe = os.path.splitext(fn)
        readers = ReaderManager.gets(fe[1:].lower(), tag=self.tag)
        if len(readers)==0: 
            return self.app.alert('no reader found for %s file'%fe[1:])
        if not self.tag is None:
            self.app.show(self.tag, readers[0][1](para['path']), fn)
        else: self.app.show(readers[0][2], readers[0][1](para['path']), fn)

class ImageWriter(Simple):
    tag = 'img'
    note = ['all']
    para={'path':''}

    def show(self):
        filt = [i.lower() for i in self.filt]
        self.para['path'] = self.app.get_path('Save..', filt, 'save', '')
        return not self.para['path'] is None

    #process
    def run(self, ips, imgs, para = None):
        fp, fn = os.path.split(para['path'])
        fn, fe = os.path.splitext(fn)
        writer = WriterManager.gets(fe[1:].lower(), tag=self.tag)
        if len(writer)==1: writer[0][1](para['path'], ips.img if self.tag=='img' else imgs)

class TableWriter(Table):
    tag = 'tab'
    note = ['all']
    para={'path':''}

    def show(self):
        filt = [i.lower() for i in self.filt]
        self.para['path'] = self.app.get_path('Save..', filt, 'save', '')
        return not self.para['path'] is None

    #process
    def run(self, tps, snap, data, para = None):
        fp, fn = os.path.split(para['path'])
        fn, fe = os.path.splitext(fn)

        writer = WriterManager.gets(fe[1:], tag=self.tag)
        if len(writer)==1: return writer[0][1](para['path'], data)


from tkinter import ttk
import tkinter as tk
from scitk.widgets.toolbar import make_logo
class ControlWidget():
    """
    parent
    widget_name
    callback
    widget_text
    """
    def __init__(self, parent: ttk.Frame, widget_name: str, callback: callable, widget_text='',widget_logo=None):
        if widget_text == '':
            self.text = self.name = widget_name
        else:
            self.name = widget_name
            self.text = widget_text
        self.widget_logo = widget_logo
        self.callback = callback
    def on_action(self):
        self.callback()# 不用显式传入self，因为自身调用的时候就已经传入了。

class ButtonWidget(ControlWidget):
    def __init__(self, parent: ttk.Frame, widget_name: str, callback: str, widget_text='',widget_logo=None):
        super().__init__(parent,widget_name,callback,widget_text,widget_logo)
        self.button = ttk.Button(parent, text=widget_text, command=self.on_action)
        self.button.pack()

class SwitchWidget(ControlWidget): # 有True和False两种状态。
    '''
    按下时为True,弹起时为False
    parent
    widget_name
    callback:tuple型，两元素。分别为按下时事件、弹起时事件。
    widget_text:tuple型，两元素。分别为按下时显示的文字，弹起状态下的文字。
    widget_logo:tuple型，两元素。分别为按下和弹起状态下显示的图片。
    '''
    def __init__(self, parent: ttk.Frame, widget_name: str, callback: str, widget_text=('on','off'),widget_logo=None,status=False):
        super().__init__(parent,widget_name,callback,widget_text,widget_logo)
        text = widget_text[1] if status==False else widget_text[0]
        self.status = status
        self.switch = ttk.Button(parent, text=text, command=self.on_action)
        self.switch.pack()
        if widget_logo is not None:
            self.widget_logo = (make_logo(widget_logo[0],(16*16)),make_logo(widget_logo[1],(16*16)))
        else:
            self.widget_logo=None

    def on_action(self):
        print('on_action',self.status)
        if self.status == False:
            self.switch.config(text=self.text[0])
            if self.widget_logo==None:
                self.switch.state(['pressed'])
            else:
                self.switch.config(image=self.widget_logo[0])
        else:
            self.switch.config(text=self.text[1])
            if self.widget_logo==None:
                self.switch.state(['!pressed'])
            else:
                self.switch.config(image=self.widget_logo[1])
        self.switch.update()
        self.status = not self.status
        self.callback()

    def configs(self):
        pass



from .action import SciAction

class TabAction(SciAction):
    title = 'Table Action'
    note, para, view = [], None, None

    def __init__(self): pass

    def show(self):
        tps, data, snap = self.tps, self.tps.data, self.tps.snap
        f = lambda p: self.run(tps, data, snap, p) or tps.update()
        return self.app.show_para(self.title, self.para, self.view, f, on_ok=None, 
            on_cancel=lambda x=self.tps:self.cancel(x), 
            preview='preview' in self.note, modal=True)

    def cancel(self, tps):
        tps.data[:] = pts.snap
        tps.update()

    def run(self, tps, snap, data, para = None):
        print('I am running!!!')

    def start(self, app, para=None, callback=None):
        self.app, self.tps = app, app.get_table()
        if 'auto_snap' in self.note:
            if 'auto_msk' in self.note: mode = True
            elif 'msk_not' in self.note: mode = False
            else: mode = None
            self.tps.snapshot(mode, 'num_only' in self.note)
        if para!=None:
            self.ok(self.tps, para, callback)
        elif self.view==None:
            if not self.__class__.show is Table.show:
                if self.show():
                    self.run(self.tps, para, callback)
            else: self.ok(self.tps, para, callback)
        elif self.modal:
            if self.show():
                self.ok(self.tps, para, callback)
            else:self.cancel(self.tps)
            if not self.dialog is None: self.dialog.Destroy()
        else: self.show()


from tkinter import ttk
import tkinter as tk
import pandas as pd
import numpy as np
class ScrollableTreeviewFrame(ttk.Frame):
    def __init__(self,master):
        super().__init__(master)
        
        self.hFrame=ttk.Frame(self)
        
        self.hFrame.pack(expand = True, fill = tk.BOTH)
        self.start=False
        self.button=ttk.Button(self,command=lambda : self.set_data(pd.DataFrame(np.random.randn(100,5))))
        self.button.pack(side=tk.LEFT)
        self.set_widgets()
        self.pack(side=tk.LEFT,expand = True, fill = tk.BOTH)

    def set_widgets(self,columns=None):
        if self.start:# 反正不好删改，就直接把控件删掉了。
            self.treeview.destroy()
            self.verticalScrollbar.destroy()
            self.horizontalScrollbar.destroy()
        self.start=True
        self.treeview = ttk.Treeview(self.hFrame, show = "headings", columns = columns, selectmode = tk.BROWSE)
        self.verticalScrollbar = ttk.Scrollbar(self.hFrame, orient="vertical", command=self.treeview.yview)
        
        self.horizontalScrollbar=ttk.Scrollbar(self,orient='horizontal',command=self.treeview.xview)
        self.treeview.pack(side=tk.LEFT,expand = True, fill = tk.BOTH)
        self.verticalScrollbar.pack(side=tk.LEFT, expand=True,fill=tk.Y)
        
        self.horizontalScrollbar.pack(expand=True,fill=tk.X)
        self.treeview.configure(yscrollcommand=self.verticalScrollbar.set)
        self.treeview.configure(xscrollcommand=self.horizontalScrollbar.set)


import tkinter as tk
from tkinter import ttk

class ClosableNotebook(ttk.Notebook):
    """A ttk Notebook with close buttons on each tab"""
    __initialized = False

    def __init__(self, *args, **kwargs):
        if not self.__initialized:
            self.__initialize_custom_style()
            self.__inititialized = True
        self.tabList=[]# 类型为ttk.Frame
        kwargs["style"] = "CustomNotebook"
        ttk.Notebook.__init__(self, *args, **kwargs)

        self._active = None

        self.bind("<ButtonPress-1>", self.on_close_press, True)
        self.bind("<ButtonRelease-1>", self.on_close_release)

    def get_tab(self,i)->ttk.Frame:
        if not i is None:
            return self.tabList[i]
        else: return self.get_current_tab()
    
    def get_current_tab(self)->ttk.Frame:
        return self.tabList[self.index("current")]
        
    def add_tab(self,tab:ttk.Frame,text:str)->None:
        self.add(tab,text=text)
        self.tabList.append(tab)

    def delete_tab(self,tab_id)->None:
        if type(tab_id)==type(""):
            if tab_id=="current":
                i=self.index("current")
            elif tab_id=='end':
                i=len(self.tabs())-1
            else:
                raise Exception("Invalid tab index:%s"%tab_id)
        else:
            i=tab_id
        self.tabList.pop(i)
        self.forget(tab_id)
        

    def on_close_press(self, event):
        """Called when the button is pressed over the close button"""

        element = self.identify(event.x, event.y)

        if "close" in element:
            index = self.index("@%d,%d" % (event.x, event.y))
            self.state(['pressed'])
            self._active = index

    def on_close_release(self, event):
        """Called when the button is released over the close button"""
        if not self.instate(['pressed']):
            return

        element =  self.identify(event.x, event.y)
        index = self.index("@%d,%d" % (event.x, event.y))

        if "close" in element and self._active == index:
            self.forget(index)
            self.event_generate("<<NotebookTabClosed>>")

        self.state(["!pressed"])
        self._active = None

    def __initialize_custom_style(self):
        style = ttk.Style()
        self.images = (
            tk.PhotoImage("img_close", data='''
            R0lGODlhCAAIAMIBAAAAADs7O4+Pj9nZ2Ts7Ozs7Ozs7Ozs7OyH+EUNyZWF0ZWQg
            d2l0aCBHSU1QACH5BAEKAAQALAAAAAAIAAgAAAMVGDBEA0qNJyGw7AmxmuaZhWEU
            5kEJADs=
            '''),
            tk.PhotoImage("img_closeactive", data='''
            R0lGODlhCAAIAMIEAAAAAP/SAP/bNNnZ2cbGxsbGxsbGxsbGxiH5BAEKAAQALAAA
            AAAIAAgAAAMVGDBEA0qNJyGw7AmxmuaZhWEU5kEJADs=
            '''),
            tk.PhotoImage("img_closepressed", data='''
            R0lGODlhCAAIAMIEAAAAAOUqKv9mZtnZ2Ts7Ozs7Ozs7Ozs7OyH+EUNyZWF0ZWQg
            d2l0aCBHSU1QACH5BAEKAAQALAAAAAAIAAgAAAMVGDBEA0qNJyGw7AmxmuaZhWEU
            5kEJADs=
        ''')
    )

        style.element_create("close", "image", "img_close",
                        ("active", "pressed", "!disabled", "img_closepressed"),
                        ("active", "!disabled", "img_closeactive"), border=8, sticky='')
        style.layout("CustomNotebook", [("CustomNotebook.client", {"sticky": "nswe"})])
        style.layout("CustomNotebook.Tab", [
        ("CustomNotebook.tab", {
            "sticky": "nswe", 
            "children": [
                ("CustomNotebook.padding", {
                    "side": "top", 
                    "sticky": "nswe",
                    "children": [
                        ("CustomNotebook.focus", {
                            "side": "top", 
                            "sticky": "nswe",
                            "children": [
                                ("CustomNotebook.label", {"side": "left", "sticky": ''}),
                                ("CustomNotebook.close", {"side": "left", "sticky": ''}),
                            ]
                    })
                ]
            })
        ]
    })
])




# -*- coding: utf-8 -*-
"""
Created on Sat Dec  3 03:32:05 2016
@author: yxl
"""
import threading
from time import time
import numpy as np

class Simple:
    title = 'SimpleFilter'
    asyn = True
    note = []
    para = None
    'all, 8-bit, 16-bit, rgb, float, req_roi, stack, stack2d, stack3d, preview'
    view = None
    prgs = None
    modal = True

    def __init__(self): pass

    def load(self, ips):return True
        
    def check(self, ips):
        note = self.note
        if ips == None:
            self.app.alert('No image opened!')
            return False
        if 'req_roi' in note and ips.roi == None:
            self.app.alert('No Roi found!')
            return False
        if not 'all' in note:
            if ips.dtype==np.uint8 and ips.channels==3 and not 'rgb' in note:
                self.app.alert('Do not surport rgb image')
                return False
            elif ips.dtype==np.uint8 and ips.channels==1 and not '8-bit' in note:
                self.app.alert('Do not surport 8-bit image')
                return False
            elif ips.dtype==np.uint16 and not '16-bit' in note:
                self.app.alert('Do not surport 16-bit uint image')
                return False
            elif ips.dtype==np.int32 and not 'int' in note:
                self.app.alert('Do not surport 32-bit int uint image')
                return False
            elif ips.dtype in {np.float32, np.float64} and not 'float' in note:
                self.app.alert('Do not surport float image')
                return False
        if sum([i in note for i in ('stack','stack2d','stack3d')])>0:
            if ips.slices==1:
                self.app.alert('Stack required!')
                return False
            elif 'stack2d' in note and ips.isarray:
                self.app.alert('Stack2d required!')
                return False
            elif 'stack3d' in note and not ips.isarray:
                self.app.alert('Stack3d required!')
                return False
        return True

    def preview(self, ips, para):pass

    def progress(self, i, n): self.prgs = int(i*100/n)

    def show(self):
        preview = lambda para, ips=self.ips: self.preview(ips, para) or ips.update()
        return self.app.show_para(self.title, self.para, self.view, preview, 
            on_ok=lambda : self.ok(self.ips), on_help=self.on_help,
            on_cancel=lambda : self.cancel(self.ips) or self.ips.update(), 
            preview='preview' in self.note, modal=self.modal)
    
    def run(self, ips, imgs, para = None):pass

    def cancel(self, ips):pass

    def on_help(self):
        self.app.show_md(self.__doc__ or 'No Document!', self.title)

    def ok(self, ips, para=None, callafter=None):
        if para == None: para = self.para
        if self.asyn and self.app.asyn:
            threading.Thread(target = self.runasyn, 
                    args = (ips, ips.imgs, para, callafter)).start()
        else: self.runasyn(ips, ips.imgs, para, callafter)

    def runasyn(self,  ips, imgs, para = None, callback = None):
        self.app.record_macros('{}>{}'.format(self.title, para))
        self.app.add_task(self)
        start = time()
        self.run(ips, imgs, para)
        self.app.info('%s: cost %.3fs'%(self.title, time()-start))
        ips.update()
        self.app.remove_task(self)
        if callback!=None:callback()
        
    def start(self, app, para=None, callback=None):
        self.app, self.ips = app, app.get_img()
        if not self.check(self.ips):return
        if not self.load(self.ips):return

        if para!=None:
            self.ok(self.ips, para, callback)
        elif self.view==None:
            if not self.__class__.show is Simple.show:
                if self.show():
                    self.ok(self.ips, para, callback)
            else: self.ok(self.ips, para, callback)
        elif self.modal:
            if self.show():
                self.ok(self.ips, para, callback)
            else:
                self.cancel(self.ips)
                self.ips.update()
        else: self.show()


class Input():# 从文本框，数据库等中输入数据。
    pass

class Drive():# 通过按钮、定时器等东西驱动的事件
    pass

class Process():# 处理过程，其中调用了控件进行设置。
    pass

class View():# 通过画布、表格等内容进行显示。
    pass

class Output(): # 将数据存储到数据库或者文件中。
    pass

class Drive():
    pass

x=r'''
{'type' : 'input',
    'name': 'input_1',
  'source' : 'xls',
  'ports' : { 
                'output' : [{'type':'pandas.DataFrame' , 'name':'output'}],
                'input' : []
                 }
 }
'''
y = r'''
{  'type':'process',
    'name': 'process_1',
    'func':'process_callback',
    'ports':{
                'input':[{'name':'input', 'type':'pandas.DataFrame'}]
                'output': [{'name':'output','type':'pandas.DataFrame'}]
                }
    'control': {(int, 'age', (0,150), 0, 'age', 'years old'),# 参数名是'age'。通过这里就可以获取了。
                    (float, 'h', (0.3, 2.5), 2, 'height', 'm')   # 运行时，首先获取这个字典中对应的值，然后用**kwargs传进去。
                    }
}
'''

if __name__=='__main__':
    d=eval(x)
    print(d)


import numpy as np

default_lut = np.arange(256*3, dtype=np.uint8).reshape((3,-1)).T

def get_updown(imgs, slices='all', chans='all', step=1):
    c = chans if isinstance(chans, int) else slice(None)
    if isinstance(slices, int): imgs = [imgs[slices]]
    if step<=1: step = int(1/step+0.5)
    else: step = int(min(imgs[0].shape[:2])/step+0.5)
    s = slice(None, None, max(step, 1))
    s = (s,s,c)[:imgs[0].ndim]
    mins = [i[s].min(axis=(0,1)) for i in imgs]
    maxs = [i[s].max(axis=(0,1)) for i in imgs]
    mins = np.array(mins).reshape((len(mins),-1))
    maxs = np.array(maxs).reshape((len(maxs),-1))
    mins, maxs = mins.min(axis=0), maxs.max(axis=0)
    if np.iscomplexobj(mins):
        mins, maxs = np.zeros(mins.shape), np.abs(maxs)
    if chans!='all': return mins.min(), maxs.max()
    return [(i,j) for i,j in zip(mins, maxs)]

def lookup(img, cn, rgs, lut):
    if isinstance(cn, int): cn = [cn]
    img = img.reshape(img.shape[:2]+(-1,))
    buf = np.zeros(img.shape[:2]+(len(cn),), dtype=np.uint8)
    for i in range(len(cn)):
        rg = rgs[cn[i]]
        k = 255.0/(max(1e-10, rg[1]-rg[0]))
        bf = np.clip(img[:,:,cn[i]], rg[0], rg[1])
        np.subtract(bf, rg[0], out=bf, casting='unsafe')
        np.multiply(bf, k, out=buf[:,:,i], casting='unsafe')
    return buf if len(cn)==3 else lut[buf.reshape(img.shape[:2])]

def histogram(imgs, rg=(0,256), slices='all', chans='all', step=1):
    c = chans if isinstance(chans, int) else slice(None)
    if isinstance(slices, int): imgs = [imgs[slices]]
    if step<=1: step = int(1/step+0.5)
    else: step = int(min(imgs[0].shape[:2])/step+0.5)
    s = slice(None, None, max(step,1))
    s = (s,s,c)[:imgs[0].ndim]
    rg = np.linspace(rg[0], rg[1], 257)
    hist = [np.histogram(i[s], rg)[0] for i in imgs]
    return np.sum(hist, axis=0)

class Image:
    def __init__(self, imgs=None, name='Image'):
        self.name = name
        self.cur = 0
        self.set_imgs(imgs)
        self.roi = None
        self.mark = None
        self.unit = 1, 'pix'
        self.msk = None
        self.pos = (0,0)
        self.cn = 0
        self.rg = (0,255)
        self.lut = default_lut
        self.log = False
        self.mode = 'set'
        self.dirty = False
        self.snap = None
        self.back = None

    @property
    def box(self):
        (x, y), (h, w) = self.pos, self.shape
        return [x, y, x+w, y+h]
    
    @property
    def title(self): return self.name
    
    @property
    def img(self): return self.imgs[self.cur]

    @img.setter
    def img(self, value): 
        self.imgs[self.cur] = value
        self.reset()
        
    def set_imgs(self, imgs):
        self.imgs = [imgs] if imgs is None else imgs
        if not imgs is None: self.reset()

    @property
    def channels(self):
        if self.img.ndim==2: return 1
        else: return self.img.shape[2]

    @property
    def slices(self): return len(self.imgs)

    @property
    def isarray(self): return isinstance(self.imgs, np.ndarray)

    @property
    def nbytes(self):
        return sum([i.nbytes for i in self.imgs])

    @property
    def dtype(self): return self.img.dtype

    @property
    def shape(self): return self.img.shape[:2]

    @property
    def info(self):
        return '%s  %sx%s  S:%s/%s  C:%s/%s  %.2fM'%(str(self.dtype).upper(), *self.shape,
            self.cur+1, self.slices, self.cn, self.channels, self.nbytes/1024/1024)

    @property
    def range(self):
        rg = np.array(self.rg).reshape((-1,2))
        return (rg[:,0].min(), rg[:,1].max())

    @range.setter
    def range(self, value):
        self.rg = [value] * len(self.rg)

    def mask(self, mode='in'):
        if self.roi==None: return None
        if self.roi.msk != mode:
            self.msk = self.roi.to_mask(self.shape, mode)
        return self.msk    

    @property
    def rect(self):
        if self.roi is None: return slice(None), slice(None)
        box, shape = self.roi.box, self.shape
        l, r = max(0, int(box[0])), min(shape[1], int(box[2]))
        t, b = max(0, int(box[1])), min(shape[0], int(box[3]))
        return slice(t,b), slice(l,r)

    def subimg(self, s1=None, s2=None):
        s1 = s1 or self.rect[0]
        s2 = s2 or self.rect[1]
        if self.isarray: return self.imgs[s1, s2]
        else: return [i[s1, s2] for i in self.imgs]

    def update(self): self.dirty = True

    def reset(self):
        self.cn = [0, (0,1,2)][self.channels==3]
        if self.dtype == np.uint8: pass
            # self.rg = [(0, 255)] * self.channels
        else: 
            self.rg = self.get_updown('all', 'all', step=512)

    def snapshot(self):
        dif = self.snap is None
        dif = dif or self.snap.shape != self.img.shape
        dif = dif or self.snap.dtype != self.img.dtype
        if dif: self.snap = self.img.copy()
        else: self.snap[:] = self.img

    def swap(self):
        if self.snap is None:return
        buf = self.img.copy()
        self.img[:], self.snap[:] = self.snap, buf

    def histogram(self, rg=None, slices=None, chans=None, step=1):
        if slices is None: slices = self.cur
        if chans is None: chans = self.cn
        if rg is None: rg = self.range
        return histogram(self.imgs, rg, slices, chans, step)

    def get_updown(self, slices='all', chans='all', step=512):
        if slices is None: slices = self.cur
        if chans is None: chans = self.cn
        return get_updown(self.imgs, slices, chans, step)

    def lookup(self, img=None):
        if img is None: img = self.img
        return lookup(img, self.cn, [self.rg], self.lut)

if __name__ == '__main__':
    img = Image(np.zeros((5,5)))



from .action import SciAction
from .imgact import ImgAction
from .tolact import Tool, DefaultTool, ImageTool, ShapeTool, TableTool
from .plugin.mea_tools import *
from .plugin.shp_tools import *
from .plugin.roi_tools import *
from .advanced import Filter, Free, Simple, Table, Macros, Widget, dataio


from .action import SciAction

class Tool(SciAction):
    title = 'Base Tool'
    default = None
    cursor = 'arrow'
    view = None
    para = None

    def config(self): pass
    def mouse_down(self, canvas, x, y, btn, **key): pass
    def mouse_up(self, canvas, x, y, btn, **key): pass
    def mouse_move(self, canvas, x, y, btn, **key): pass
    def mouse_wheel(self, canvas, x, y, d, **key): pass
    def start(self, app, para=None, callafter=None): 
        self.app, self.default = app, self
        if not app is None: app.tool = self

class DefaultTool(Tool):
    title = 'Move And Scale'
    def __init__(self): 
        self.oldxy = None

    def mouse_down(self, obj, x, y, btn, **key):
        if btn==1: self.oldxy = key['px'], key['py']
        if btn==3: key['canvas'].fit()
        
    def mouse_up(self, obj, x, y, btn, **key):
        self.oldxy = None
    
    def mouse_move(self, obj, x, y, btn, **key):
        if self.oldxy is None: return
        ox, oy = self.oldxy
        up = (1,-1)[key['canvas'].up]
        key['canvas'].move(key['px']-ox, (key['py']-oy)*up)
        self.oldxy = key['px'], key['py']
    
    def mouse_wheel(self, obj, x, y, d, **key):
        if d>0: key['canvas'].zoomout(x, y, coord='data')
        if d<0: key['canvas'].zoomin(x, y, coord='data')

    def start(self, app, para=None): 
        self.app = app
        Tool.default = self
        if not app is None: app.tool = self

class ImageTool(DefaultTool):
    default = None
    title = 'Image Tool'

    def mouse_move(self, img, x, y, btn, **key):
        if self.app is None: return
        r, c = int(y), int(x)
        if (r>0) & (c>0) & (r<img.shape[0]) & (c<img.shape[1]):
            s = 'x:%d y:%d  value:%s'%(x, y, img.img[r,c])
            self.app.info(s)

    def start(self, app, para=None, callafter=None): 
        self.app = app
        ImageTool.default = self
        if not app is None: app.tool = self

class ShapeTool(DefaultTool):
    default = None
    title = 'Shape Tool'

    def mouse_move(self, img, x, y, btn, **key):
        if self.app: self.app.info('%d, %d'%(x, y))

    def start(self, app, para=None, callafter=None): 
        self.app = app
        ShapeTool.default = self
        if not app is None: app.tool = self

class TableTool(DefaultTool):
    default = None
    title = 'Table Tool'

    def mouse_down(self, data, x, y, btn, **others):
        print('you click on cell', x, y)

    def start(self, app, para=None, callafter=None): 
        self.app = app
        TableTool.default = self
        if not app is None: app.tool = self

DefaultTool().start(None)
ImageTool().start(None)
ShapeTool().start(None)
TableTool().start(None)


import tkinter as tk
from tkinter import ttk
import sys
import numpy as np
from numpy.linalg import norm
from scipy import interpolate

class CurvePanel(ttk.Frame):
    """ HistCanvas: diverid from wx.core.Panel """
    def __init__(self, parent, hist=None, l=255, app=None):
        super().__init__ (parent)
        self.init_buf()
        self.offset = (20, 5)
        self.l, self.k = l, l/255.0
        self.idx = -1
        self.set_hist(hist)
        self.dirty = False
        self.pts = [(0,0), (255, 255)]
        wx.Panel.Bind(self, wx.EVT_SIZE, self.on_size)  
        wx.Panel.Bind(self, wx.EVT_IDLE, self.on_idle)
        wx.Panel.Bind(self, wx.EVT_PAINT, self.on_paint)
        wx.Panel.Bind(self, wx.EVT_LEFT_DOWN, self.on_ld)
        wx.Panel.Bind(self, wx.EVT_LEFT_UP, self.on_lu )
        wx.Panel.Bind(self, wx.EVT_MOTION, self.on_mv )
        wx.Panel.Bind(self, wx.EVT_RIGHT_DOWN, self.on_rd )

    def update(self): self.dirty = True
    
    @classmethod
    def lookup(cls, pts):
        x, y = np.array(pts).T
        kind = 'linear' if len(pts)==2 else 'quadratic'
        f = interpolate.interp1d(x, y, kind=kind)
        return np.clip(f(np.arange(256)),0,255).astype(np.uint8)
            
    def init_buf(self):
        box = self.GetClientSize()
        if min(box)==0: return
        self.buffer = wx.Bitmap(box.width, box.height)
        
    def on_size(self, event):
        self.init_buf()
        self.update()
        
    def on_idle(self, event):
        if self.dirty == True:
            self.draw()
            self.dirty = False

    def pick(self, x, y):
        dis = norm(np.array(self.pts)-(x,y), axis=1)
        if dis[np.argmin(dis)] > 3: return -1
        return np.argmin(dis)

    def on_ld(self, event):
        x = (event.GetX()-self.offset[0])/self.k 
        y = (event.GetY()-self.offset[1])/self.k
        self.idx = self.pick(x, 255-y)
        if self.idx==-1: 
            self.pts.append((x, 255-y))
            self.idx = len(self.pts)-1
            self.update()
            self.on_curve(self.GetValue())

    def on_lu(self, event):
        self.idx = -1

    def on_rd(self, event):
        x = (event.GetX()-self.offset[0])/self.k 
        y = (event.GetY()-self.offset[1])/self.k
        self.idx = self.pick(x, 255-y)
        if self.idx==-1: return
        if not self.pts[self.idx][0] in (0, 255):
            del self.pts[self.idx]
            self.idx = -1
            self.update()
            self.on_curve(self.GetValue())

    def on_mv(self, event):
        x = (event.GetX()-self.offset[0])/self.k 
        y = (event.GetY()-self.offset[1])/self.k
        if self.pick(x, 255-y)!=-1:
            self.SetCursor(wx.Cursor(wx.CURSOR_HAND))
        else: self.SetCursor(wx.Cursor(wx.CURSOR_ARROW))
        if self.idx!=-1:
            oldx = self.pts[self.idx][0]
            if oldx == 0: x=0
            elif oldx==255: x=255
            else: x = np.clip(x, 1, 254)
            y = np.clip(y, 0, 255)
            self.pts[self.idx] = (x, 255-y)
            self.update()
            self.on_curve(self.GetValue())
        

    def on_paint(self, event):
        wx.BufferedPaintDC(self, self.buffer)
        
    def set_hist(self, hist):
        if hist is None:self.hist=None
        else:
            self.hist = (hist*self.l/hist.max())
            self.logh = (np.log(self.hist+1.0))*(self.l/(np.log(self.l+1)))
        self.update()
        
    def set_pts(self, pts):
        self.x1, self.x2 = x1, x2
        self.update()

    def draw(self):
        ox, oy = self.offset
        # get client device context buffer
        dc = wx.BufferedDC(wx.ClientDC(self), self.buffer)
        dc.Clear()
        # w, h = self.GetClientSize()
        
        # the main draw process 
        
        if not self.hist is None:      
            dc.SetPen(wx.Pen((200,200,200), width=1, style=wx.SOLID))
            for i in np.linspace(0,self.l,256).astype(np.int16):
                dc.DrawLine(i+ox,self.l+1+oy,i+ox,self.l+1-self.logh[i]+oy)
            dc.SetPen(wx.Pen((100,100,100), width=1, style=wx.SOLID))
            for i in np.linspace(0,self.l,256).astype(np.int16):
                dc.DrawLine(i+ox,self.l+1+oy,i+ox,self.l+1-self.hist[i]+oy)
        x, y = np.array(self.pts).T
        kind = 'linear' if len(self.pts)==2 else 'quadratic'
        f = interpolate.interp1d(x, y, kind=kind)
        ys = np.clip(f(np.linspace(0, 255, self.l+1)), 0, 255)
        if ys is None:return
        dc.SetBrush(wx.Brush((0,0,0)))
        dc.SetPen(wx.Pen((0,0,0), width=1))
        for i in self.pts: 
            dc.DrawCircle(round(i[0]*self.k+ox), round(self.l+1-i[1]*self.k+oy), 2)
        xs, ys = np.linspace(0,self.l, self.l+1)+ox, self.l+1-ys*self.k+oy
        dc.DrawPointList(list(zip(xs.round(), ys.round())))

        dc.SetPen(wx.Pen((0,0,0), width=1, style=wx.SOLID))
        for i in np.linspace(0, self.l+1, 5):
            dc.DrawLine(0+ox, i+oy, self.l+1+ox, i+oy)
            dc.DrawLine(i+ox, 0+oy, i+ox, self.l+1+oy)
        dc.SetBrush(wx.Brush((0,0,0), wx.BRUSHSTYLE_TRANSPARENT))
        arr = np.zeros((10,self.l+1,3),dtype=np.uint8)
        arr[:] = np.vstack([np.linspace(0,255,self.l+1)]*3).T
        bmp = wx.Bitmap.FromBuffer(self.l+1,10, memoryview(arr))
        dc.DrawBitmap(bmp, 0+ox, self.l+6+oy)
        dc.DrawRectangle(0+ox, self.l+6+oy, self.l+1, 10)
        arr = arr.transpose((1,2,0))[::-1].copy()
        bmp = wx.Bitmap.FromBuffer(10, self.l+1, memoryview(arr))
        dc.DrawBitmap(bmp, -15+ox, 0+oy)
        dc.DrawRectangle(-15+ox, 0+oy, 10, self.l+1)
        
    def on_curve(self, event): print(event)
    
    def Bind(self, event, f):self.on_curve = f

    def SetValue(self, value=None):
        print('here', value)
        if not value is None:
            self.pts = value
        else: self.pts = [(0,0), (255, 255)]
        self.update()

    def GetValue(self): return sorted(self.pts)

if __name__ == '__main__':
    app = wx.App()
    frame = wx.Frame(None)
    hist = CurvePanel(frame, l=255)
    hist.set_hist(np.random.rand(256)+2)
    frame.Fit()
    frame.Show(True)
    app.MainLoop() 



from .action import SciAction

class ImgAction(SciAction):
    title = 'Image Action'
    note, para, view = [], None, None

    def __init__(self): pass

    def show(self):
        ips, img, snap = self.ips, self.ips.img, self.ips.snap
        f = lambda p: self.run(ips, img, snap, p) or self.ips.update()
        return self.app.show_para(self.title, self.para, self.view, f, on_ok=None, 
            on_cancel=lambda x=self.ips:self.cancel(x), 
            preview='preview' in self.note, modal=True)

    def cancel(self, ips):
        ips.img[:] = ips.snap
        ips.update()

    def run(self, ips, img, snap, para):
        print('I am running!!!')

    def start(self, app, para=None, callback=None):
        print('Image Action Started!')
        self.app = app
        self.ips = app.get_img()
        if 'auto_snap' in self.note: self.ips.snapshot()
        if para!=None:
            self.run(self.ips, self.ips.img, self.ips.snap, para)
        elif self.view==None and self.__class__.show is ImgAction.show:
            self.run(self.ips, self.ips.img, self.ips.snap, para)
        elif self.show():
            self.run(self.ips, self.ips.img, self.ips.snap, self.para)
        elif 'auto_snap' in self.note: self.cancel(self.ips)
        self.ips.update()



from .manager import Manager
from .object import Image, Table

class App():
    def __init__(self, asyn=True):
        self.asyn = asyn
        self.managers = {}
        self.img_manager = self.manager('img')
        #self.wimg_manager = self.manager('wimg')
        self.tab_manager = self.manager('tab')
        #self.wtab_manager = self.manager('wtab')
        self.mesh_manager = self.manager('mesh')
        #self.wmesh_manager = self.manager('wmesh')
        self.task_manager = self.manager('task')
        
        self.plugin_manager = self.manager('plugin')

    def manager(self, name, value=None):
        if not name in self.managers: 
            self.managers[name] = Manager()
        return self.managers[name]

    # ========== Plugin ==========
    def add_plugin(self, name, plg, tag=None):
        self.plugin_manager.add(name, plg, tag)
        for i in ' _.-': name = name.replace(i,'_')
        self.__dict__['_%s_'%name] = plg

    def get_plugin(self, name=None):
        return self.plugin_manager.get(name)

    def plugin_names(self, tag=None):
        return self.plugin_manager.names(tag)

    # ========== Image ==========
    def show_img(self, img, name): 
        if not isinstance(img, Image): 
            img = Image(img, name)
        img.name = self.img_manager.name(name)
        self.img_manager.add(name, img)
        print(img.info)

    def close_img(self, name): 
        self.img_manager.remove(name)
        print('close image:', name)

    def active_img(self, name): 
        self.img_manager.active(name)
        print('active image:', name)

    def get_img(self, name=None):
        return self.img_manager.get(name)

    def img_names(self):
        return self.img_manager.names()

    # ========== Table ==========
    def show_table(self, tab, name): 
        if not isinstance(tab, Table): 
            tab = Table(tab, name)
        tab.name = self.tab_manager.name(name)
        self.tab_manager.add(name, tab)
        print(tab.info)

    def close_table(self, name): 
        self.tab_manager.remove(name)
        print('close table:', name)

    def active_table(self, name): 
        self.tab_manager.active(name)
        print('active image:', name)

    def get_table(self, name=None):
        return self.tab_manager.get(name)

    def table_names(self):
        return self.tab_manager.names()
    # ========== Others ==========
    def show_plot(self): pass

    def show_mesh(self): pass

    def add_mesh(self, mesh):
        if not self.mesh_manager.has(mesh.name, obj=mesh):
            mesh.name = self.mesh_manager.name(mesh.name)
        self.mesh_manager.add(mesh.name, mesh)

    def remove_mesh(self, mesh):
        self.mesh_manager.remove(obj=mesh)

    def add_mesh_win(self, win):
        self.wmesh_manager.add(win.name, win)

    def remove_mesh_win(self, win):
        self.wmesh_manager.remove(obj=win)

    def get_mesh(self, name=None):
        return self.mesh_manager.get(name)
    
    def get_mesh_name(self):
        return self.mesh_manager.names()
    
    def get_mesh_win(self, name=None):
        return self.wmesh_manager.get(name)

    def add_task(self, task):
        self.task_manager.add(task.title, task)

    def remove_task(self, task):
        self.task_manager.remove(obj=task)

    def info(self, value): 
        print('Information:', value)

    def show_md(self, cont, title='ImagePy'):
        print(title, '\n', cont)

    def show_txt(self, cont, title='ImagePy'):
        print(title, '\n', cont)

    def alert(self, cont, title='ImagePy'):
        print(title, '\n', cont, 'enter to continue!')
        input()

    def yes_no(self, cont, title='ImagePy'):
        print(title, '\n', cont, 'Y/N?')
        return input() in 'yY'

    def get_path(self, title='sciapp', filt='', io='save', name=''):
        print('input file path:')
        return input()
    
    def show_para(self, title, para, view, on_handle=None, on_ok=None, 
        on_cancel=None, on_help=None, preview=False, modal=True):
        for i in view:
            if i[0]==str: para[i[1]] = input(i[2]+': ? '+i[3]+' <str> ')
            if i[0]==int: para[i[1]] = int(input(i[4]+': ? '+i[5]+' <int> '))
            if i[0]==float: para[i[1]] = float(input(i[4]+': ? '+i[5]+' <float> '))
            if i[0]=='slide': para[i[1]] = float(input(i[4]+': ? <float> '))
            if i[0]==bool: para[i[1]] = bool(input(i[2]+': <True/False> '))
            if i[0]==list: para[i[1]] = i[3](input('%s %s: %s'%(i[4],i[5],i[2])+' <single choice> '))
            if i[0]=='chos':para[i[1]] = input('%s:%s <multi choices> '%(i[3],i[2])).split(',')
            if i[0]=='color': para[i[1]] = eval(input(i[2]+': ? '+i[3]+' <rgb> '))
        return para

    def run_macros(self, cmd, callafter=None):
        cmds = [i for i in cmd]
        def one(cmds, after): 
            cmd = cmds.pop(0)
            if not isinstance(cmd, str): title, para = cmd
            else: title, para = eval(cmd.replace('>', ','))
            plg = self.plugin_manager.get(name=title)()
            after = lambda cmds=cmds: one(cmds, one)
            if len(cmds)==0: after = callafter
            plg.start(self, para, after)
        one(cmds, None)

    def show(self, tag, cont, title):
        tag = tag or 'img'
        if tag=='img':
            self.show_img([cont], title)
        elif tag=='imgs':
            self.show_img(cont, title)
        elif tag=='tab':
            self.show_table(cont, title)
        elif tag=='mc':
            self.run_macros(cont)
        elif tag=='md':
            self.show_md(cont, title)
        elif tag=='wf':
            self.show_workflow(cont, title)
        else: self.alert('no view for %s!'%tag)

    def record_macros(self, cmd):
        print('>>>', cmd)


import numpy as np
from time import time
from collections import Iterable
import shapely.geometry as geom
from shapely.ops import unary_union

def merge(a, b):
    x1, y1 = min(a[0],b[0]), min(a[1],b[1])
    x2, y2 = max(a[2],b[2]), max(a[3],b[3])
    return [x1, y1, x2, y2]

class Shape:
    default = {'color':(255,255,0), 'fcolor':(255,255,255), 
    'fill':False, 'lw':1, 'tcolor':(255,0,0), 'size':8}
    dtype = 'shape'
    def __init__(self, body=None, **key):
        self.name = 'shape'
        self.body = [] if body is None else body
        self.color = key['color'] if 'color' in key else None
        self.fcolor = key['fcolor'] if 'fcolor' in key else None
        self.lstyle = key['lstyle'] if 'lstyle' in key else None
        self.tcolor = key['tcolor'] if 'tcolor' in key else None
        self.lw = key['lw'] if 'lw' in key else None
        self.r = key['r'] if 'r' in key else None
        self.fill = key['fill'] if 'fill' in key else None
        self._box = None
        self.dirty = True

    @property
    def box(self):
        if self._box is None or self.dirty:
            self._box = self.count_box()
        return self._box

    @property
    def style(self):
        styledic = {'type':self.dtype}
        if not self.color is None: styledic['color']=self.color
        if not self.fcolor is None: styledic['fcolor']=self.fcolor
        if not self.lstyle is None: styledic['lstyle']=self.lstyle
        if not self.lw is None: styledic['lw']=self.lw
        if not self.fill is None: styledic['fill']=self.fill
        return styledic
        
    def count_box(self, body=None, box=None):
        if body is None:
            box = [1e10, 1e10,-1e10,-1e10]
            self.count_box(self.body, box)
            return box
        if isinstance(body, np.ndarray):
            body = body.reshape((-1,2))
            minx, miny = body.min(axis=0)
            maxx, maxy = body.max(axis=0)
            newbox = [minx, miny, maxx, maxy]
            box.extend(merge(box, newbox))
            del box[:4]
        else:
            for i in body: self.count_box(i, box)

    @property
    def info(self):
        minx, miny, maxx, maxy = self.box
        return 'Type:%s   minX:%.3f maxX:%.3f, minY:%.3f maxY%.3f'%(
            self.dtype, minx, maxx, miny, maxy)

    def to_mark(self, body):
        mark = self.style
        mark['body'] = body
        return mark

    def to_json(self):
        return geom.mapping(self.to_geom())

    def to_geom(self): pass

    def __str__(self):
        return str(self.to_mark())
    
class Point(Shape):
    dtype = 'point'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def to_mark(self):
        return Shape.to_mark(self, tuple(self.body.tolist()))

    def to_geom(self):
        return geom.Point(self.body)

class Points(Shape):
    dtype = 'points'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def to_mark(self):
        return Shape.to_mark(self, [tuple(i.tolist()) for i in self.body])

    def to_geom(self):
        return geom.MultiPoint(self.body)

class Line(Shape):
    dtype = 'line'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def to_mark(self):
        return Shape.to_mark(self, [tuple(i.tolist()) for i in self.body])

    def to_geom(self):
        return geom.LineString(self.body)

class Lines(Shape):
    dtype = 'lines'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = [np.array(i, dtype=np.float32) for i in body]

    def to_mark(self):
        return Shape.to_mark(self, [[tuple(i.tolist()) for i in j] for j in self.body])

    def to_geom(self):
        return geom.MultiLineString(self.body)
    
class Polygon(Shape):
    dtype = 'polygon'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        if len(body)>0 and not isinstance(body[0][0], Iterable): body = [body]
        self.body = [np.array(i, dtype=np.float32) for i in body]

    def to_mark(self):
        return Shape.to_mark(self, [[tuple(i.tolist()) for i in j] for j in self.body])

    def to_geom(self):
        return geom.Polygon(self.body[0], self.body[1:])

class Polygons(Shape):
    dtype = 'polygons'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        for i in range(len(body)):
            if not isinstance(body[i][0][0], Iterable): body[i] = [body[i]]
        self.body = [[np.array(i, dtype=np.float32) for i in j] for j in body]

    def to_mark(self):
        mark = self.style
        f = lambda x:[[tuple(i.tolist()) for i in j] for j in x]
        return Shape.to_mark(self, [[[tuple(i.tolist()) for i in j] for j in k] for k in self.body])

    def to_geom(self):
        return geom.MultiPolygon([[i[0], i[1:]] for i in self.body])

class Circle(Shape):
    dtype = 'circle'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def count_box(self):
        pts = self.body
        return [pts[0]-pts[2], pts[1]-pts[2], pts[0]+pts[2], pts[1]+pts[2]]
    
    def to_mark(self):
        return Shape.to_mark(self, tuple(self.body.tolist()))

    def to_geom(self):
        return geom.Point(self.body[:2]).buffer(self.body[2])

class Circles(Shape):
    dtype = 'circles'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def count_box(self):
        pts = self.body.T
        return [(pts[0]-pts[2]).min(), (pts[1]-pts[2]).min(), (pts[0]+pts[2]).max(), (pts[1]+pts[2]).max()]
    
    def to_mark(self):
        return Shape.to_mark(self, [tuple(i.tolist()) for i in self.body])

    def to_geom(self):
        return geom.MultiPolygon([geom.Point(i[:2]).buffer(i[2]) for i in self.body])

class Rectangle(Shape):
    dtype = 'rectangle'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def count_box(self):
        return [self.body[0], self.body[1], self.body[0]+self.body[2], self.body[1]+self.body[3]]
    
    def to_mark(self):
        return Shape.to_mark(self, tuple(self.body.tolist()))

    def to_geom(self):
        f = lambda x:[(x[0],x[1]),(x[0],x[1]+x[3]),(x[0]+x[2],x[1]+x[3]),(x[0]+x[2],x[1])]
        return geom.Polygon(f(self.body))

class Rectangles(Shape):
    dtype = 'rectangles'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def count_box(self, body=None, box=None):
        self.body[:,2:] += self.body[:,:2]
        minx, miny = self.body.reshape((-1,2)).min(axis=0)
        maxx, maxy = self.body.reshape((-1,2)).max(axis=0)
        self.body[:,2:] -= self.body[:,:2]
        return [minx, miny, maxx, maxy]
    
    def to_mark(self):
        return Shape.to_mark(self, [tuple(i.tolist()) for i in self.body])

    def to_geom(self):
        f = lambda x:[(x[0],x[1]),(x[0],x[1]+x[3]),(x[0]+x[2],x[1]+x[3]),(x[0]+x[2],x[1])]
        return geom.MultiPolygon([[f(i),[]] for i in self.body])

def make_ellipse(x0, y0, l1, l2, ang):
    m = np.array([[l1*np.cos(-ang),-l2*np.sin(-ang)],
                  [l1*np.sin(-ang),l2*np.cos(-ang)]])
    a = np.linspace(0, np.pi*2, 36)
    xys = np.array((np.cos(a), np.sin(a)))
    return np.dot(m, xys).T + (x0, y0)

class Ellipse(Shape):
    dtype = 'ellipse'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def count_box(self):
        xy = make_ellipse(*self.body)
        (minx, miny), (maxx, maxy) = xy.min(axis=0), xy.max(axis=0)
        return [minx, miny, maxx, maxy]
    
    def to_mark(self):
        return Shape.to_mark(self, tuple(self.body.tolist()))

    def to_geom(self):
        return geom.Polygon(make_ellipse(*self.body))

class Ellipses(Shape):
    dtype = 'ellipses'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body, dtype=np.float32)

    def count_box(self):
        xy = np.vstack([make_ellipse(*i) for i in self.body])
        (minx, miny), (maxx, maxy) = xy.min(axis=0), xy.max(axis=0)
        return [minx, miny, maxx, maxy]
    
    def to_mark(self):
        return Shape.to_mark(self, [tuple(i.tolist()) for i in self.body])

    def to_geom(self):
        return geom.MultiPolygon([[make_ellipse(*i),[]] for i in self.body])

class Text(Shape):
    dtype = 'text'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        self.body = np.array(body[:2], dtype=np.float32)
        self.txt = body[2]

    def to_mark(self):
        return Shape.to_mark(self, tuple(self.body.tolist())+(self.txt,))

    def to_geom(self):
        return geom.Point(self.body)

class Texts(Shape):
    dtype = 'texts'
    def __init__(self, body=[], **key):
        Shape.__init__(self, body, **key)
        body = np.array(body, dtype=object)
        self.body = body[:,:2].astype(np.float32)
        self.txt = body[:,2]

    def to_mark(self):
        return Shape.to_mark(self, [(i,j,t) for (i,j), t in zip(self.body.tolist(), self.txt.tolist())])

    def to_geom(self):
        return geom.MultiPoint(self.body)
    
class Layer(Shape):
    dtype = 'layer'
    def __init__(self, body=None, **key):
        Shape.__init__(self, body, **key)

    @property
    def box(self):
        if len(self.body)==0: return None
        boxs = np.array([i.box for i in self.body]).T
        return [boxs[0].min(), boxs[1].min(), boxs[2].max(), boxs[3].max()]

    def to_mark(self):
        return Shape.to_mark(self, [i.to_mark() for i in self.body])

    def to_geom(self):
        return geom.GeometryCollection([i.to_geom() for i in self.body])

class Layers(Shape):
    dtype = 'layers'
    def __init__(self, body=None, **key):
        Shape.__init__(self, body, **key)
        
    @property
    def box(self):
        if len(self.body)==0: return None
        boxs = np.array([i.box for i in self.body.values()]).T
        return [boxs[0].min(), boxs[1].min(), boxs[2].max(), boxs[3].max()]

    def to_mark(self):
        body = dict(zip(self.body.keys(), [i.to_mark() for i in self.body.values()]))
        return Shape.to_mark(self, body)

def mark2shp(mark):
    style = mark.copy()
    style.pop('body')
    keys = {'point':Point, 'points':Points, 'line':Line, 'lines':Lines,
            'polygon':Polygon, 'polygons':Polygons, 'circle':Circle,
            'circles':Circles, 'rectangle':Rectangle, 'rectangles':Rectangles,
            'ellipse':Ellipse, 'ellipses':Ellipses, 'text':Text, 'texts':Texts}
    if mark['type'] in keys: return keys[mark['type']](mark['body'], **style)
    if mark['type']=='layer':
        return Layer([mark2shp(i) for i in mark['body']], **style)
    if mark['type']=='layers':
        return Layers(dict(zip(mark['body'].keys(),
            [mark2shp(i) for i in mark['body'].values()])), **style)

def json2shp(obj):
    if obj['type']=='Point':
        return Point(obj['coordinates'])
    if obj['type']=='MultiPoint':
        return Points(obj['coordinates'])
    if obj['type']=='LineString':
        return Line(obj['coordinates'])
    if obj['type']=='MultiLineString':
        return Lines(obj['coordinates'])
    if obj['type']=='Polygon':
        return Polygon(obj['coordinates'])
    if obj['type']=='MultiPolygon':
        return Polygons(obj['coordinates'])
    if obj['type']=='GeometryCollection':
        return Layer([json2shp(i) for i in obj['geometries']])

def geom2shp(obj): return json2shp(geom.mapping(obj))

if __name__ == '__main__':
    import json

    mark = {'type': 'layer', 'body': [{'type': 'circle', 'body': (256, 256, 5)}, {'type': 'cirlce', 'body': (256, 256, 50)}, {'type': 'circle', 'body': (306.0, 256.0, 3)}]}
    a = mark2shp(mark)
    print(a)
    #import geonumpy.io as gio
    #shp = gio.read_shp('C:/Users/54631/Documents/projects/huangqu/demo/shape/province.shp')
    #feas = json.loads(shp.to_json())['features']



from scitk.widgets.tkwidgetext.closablenotebook import ClosableNotebook
from .paradialog import ParaDialog, get_para
##from .cmappanel import CMapPanel
##from .colormap import CMapSelPanel, CMapSelCtrl
##from .curvepanel import CurvePanel
##from .histpanel import HistPanel
##from .progressbar import ProgressBar
##from .normal import *
from .toolbar import Toolbar
from .menubar import MenuBar
##from .viewport import ViewPort
##from .choicebook import ChoiceBook
##from .workflow import WorkFlowPanel
##from . import util




import tkinter as tk
from tkinter import ttk
import platform
from scitk.widgets.normal import *
##from .histpanel import HistPanel

widgets = { 'ctrl':None, 'slide':FloatSlider, int:NumCtrl, 'path':PathCtrl,
            float:NumCtrl, 'lab':Label, bool:Check, str:TextCtrl, list:Choice,
            'color':ColorCtrl, 'any':AnyType, 'chos':Choices}# 'hist':HistPanel}

def add_widget(key, value): widgets[key] = value

class ParaDialogFrame (ttk.Frame):
    def __init__( self, parent, title):
        super().__init__ (parent)
        self.lst = ttk.Frame(self)#wx.BoxSizer( wx.VERTICAL )
        self.lst.pack(expand = 1, fill=tk.BOTH)
        self.tus = []
        self.on_ok = self.on_cancel = self.on_help = None
        
        self.handle = print
        self.ctrl_dic = {}

    def commit(self, state):
        self.destroy()
        if state=='ok' and self.on_ok:self.on_ok()
        if state=='cancel' and self.on_cancel:self.on_cancel()

    def add_confirm(self, modal):
        frame = ttk.Frame(self)
        frame.pack(fill= tk.BOTH, expand = True)
        return
            
    def init_view(self, items, para, preview=False, modal = True):
        self.para, self.modal = para, modal
        for item in items:
            self.add_ctrl_(widgets[item[0]], item[1], item[2:])
        if preview:self.add_ctrl_(Check, 'preview', ('preview',))
        self.reset(para)
        self.add_confirm(modal)
        self.pack()
        print('bind close')
    
    
    def OnDestroy( self, event ):
        self.set_handle(None)
        self.on_cancel = self.on_ok = self.on_help = None
        del self.ctrl_dic

    def parse(self, para) :
        self.add_ctrl_(widgets[para[0]], *para[1:])

    def add_ctrl_(self, Ctrl, key, p):
        ctrl = Ctrl(self.lst, *p)
        if not p[0] is None: 
            self.ctrl_dic[key] = ctrl
        if hasattr(ctrl, 'ctrl'):
            pass
            ctrl.ctrl.bind('<KeyRelease>',lambda e: self.para_changed(e,ctrl), add = '+')
            ctrl.ctrl.bind('<Button>',lambda e: self.para_changed(e,ctrl), add = '+')
        if hasattr(ctrl,'on_check_callback'): # 因为事件绑定实在是不怎么样。event是直接留了一个字符串进去。
            ctrl.on_check_callback  = lambda : self.para_changed('button click',ctrl)
        pre = ctrl.prefix if hasattr(ctrl, 'prefix') else None
        post = ctrl.postfix if hasattr(ctrl, 'postfix') else None
        self.tus.append((pre, post))
        ctrl.pack(expand = True, fill=tk.X)        

##    def pack(self):
##        '''
##        这个函数是干什么的？
##        '''
##        return 
##        mint, minu = [], []
##        for t,u in self.tus:
##            if not t is None: mint.append(t.GetSize()[0])
##            if not u is None:minu.append(u.GetSize()[0])
##        for t,u in self.tus:
##            if not t is None:t.SetInitialSize((max(mint),-1))
##            if not u is None:u.SetInitialSize((max(minu),-1))

    def para_check(self, para, key):pass

    def para_changed(self,event, obj):
        key = ''
        para = self.para
        for p in self.ctrl_dic:
            if p in para:
                para[p] = self.ctrl_dic[p].GetValue()
            if self.ctrl_dic[p] == obj: key = p

        sta = sum([i is None for i in list(para.values())])==0

        if not sta: return
        self.para_check(para, key)
        if 'preview' not in self.ctrl_dic:return
        if not self.ctrl_dic['preview'].GetValue():
            if key=='preview' and self.on_cancel != None: 
                return self.on_cancel()
            else: return
        self.handle(para)

    def reset(self, para=None):
        if para!=None:self.para = para
        for p in list(self.para.keys()):
            if p in self.ctrl_dic:
                self.ctrl_dic[p].SetValue(self.para[p])

    def get_para(self): return self.para

    def Bind(self, tag, f):
        if tag == 'parameter': self.handle = f if not f is None else print
        if tag == 'commit': self.on_ok = f
        if tag == 'cancel': self.on_cancel = f

    def show(self):
        if self.modal: 
            status =  self.ShowModal() == 5100
            self.Destroy()
            return status
        else: self.Show()

    def __del__( self ):
        print('panel config deleted!')


class ParaDialog (tk.Toplevel):
    def __init__( self, parent, title):
        super().__init__ (parent)
        self.lst = ttk.Frame(self)#wx.BoxSizer( wx.VERTICAL )
        self.lst.pack(expand = 1, fill=tk.BOTH)
        self.tus = []
        self.on_ok = self.on_cancel = self.on_help = None
        
        self.handle = print
        self.ctrl_dic = {}

    def commit(self, state):
        self.destroy()
        if state=='ok' and self.on_ok:self.on_ok()
        if state=='cancel' and self.on_cancel:self.on_cancel()

    def add_confirm(self, modal):
        frame = ttk.Frame(self)
        frame.pack(fill= tk.BOTH, expand = True)
        self.btn_ok = tk.Button( frame, text =  'OK', command = self.on_ok)
        self.btn_ok.pack(side = tk.LEFT)

        self.btn_cancel = tk.Button( frame, text = 'Cancel')
        self.btn_cancel.pack(side = tk.LEFT)

        self.btn_help = tk.Button( self, text =  'Help')
        self.btn_help.pack(side = tk.LEFT)
        
        self.btn_help.config(command =  lambda : self.on_help and self.on_help())
        if not modal:
            self.btn_ok.config( command =  lambda :self.commit('ok'))
            self.btn_cancel.config( command = lambda :self.commit('cancel'))
        self.on_ok = lambda:print(self.para)
            
    def init_view(self, items, para, preview=False, modal = True):
        self.para, self.modal = para, modal
        for item in items:
            self.add_ctrl_(widgets[item[0]], item[1], item[2:])
        if preview:self.add_ctrl_(Check, 'preview', ('preview',))
        self.reset(para)
        self.add_confirm(modal)
        self.pack()
        print('bind close')
    
    
    def OnDestroy( self, event ):
        self.set_handle(None)
        self.on_cancel = self.on_ok = self.on_help = None
        del self.ctrl_dic

    def parse(self, para) :
        self.add_ctrl_(widgets[para[0]], *para[1:])

    def add_ctrl_(self, Ctrl, key, p):
        ctrl = Ctrl(self.lst, *p)
        if not p[0] is None: 
            self.ctrl_dic[key] = ctrl
        if hasattr(ctrl, 'ctrl'):
            ctrl.ctrl.bind('<KeyRelease>',lambda e: self.para_changed(e,ctrl), add = '+')
            ctrl.ctrl.bind('<Button>',lambda e: self.para_changed(e,ctrl), add = '+')
        if hasattr(ctrl,'on_check_callback'): # 因为事件绑定实在是不怎么样。event是直接留了一个字符串进去。
            ctrl.on_check_callback  = lambda : self.para_changed('button click',ctrl)
        pre = ctrl.prefix if hasattr(ctrl, 'prefix') else None
        post = ctrl.postfix if hasattr(ctrl, 'postfix') else None
        self.tus.append((pre, post))
        ctrl.pack(expand = True, fill=tk.X)        

    def pack(self):
        '''
        这个函数是干什么的？
        '''
        return 
        mint, minu = [], []
        for t,u in self.tus:
            if not t is None: mint.append(t.GetSize()[0])
            if not u is None:minu.append(u.GetSize()[0])
        for t,u in self.tus:
            if not t is None:t.SetInitialSize((max(mint),-1))
            if not u is None:u.SetInitialSize((max(minu),-1))

    def para_check(self, para, key):pass

    def para_changed(self,event, obj):
        key = ''
        para = self.para
        for p in self.ctrl_dic:
            if p in para:
                para[p] = self.ctrl_dic[p].GetValue()
            if self.ctrl_dic[p] == obj: key = p

        sta = sum([i is None for i in list(para.values())])==0

        if not sta: return
        self.para_check(para, key)
        if 'preview' not in self.ctrl_dic:return
        if not self.ctrl_dic['preview'].GetValue():
            if key=='preview' and self.on_cancel != None: 
                return self.on_cancel()
            else: return
        self.handle(para)

    def reset(self, para=None):
        if para!=None:self.para = para
        for p in list(self.para.keys()):
            if p in self.ctrl_dic:
                self.ctrl_dic[p].SetValue(self.para[p])

    def get_para(self): return self.para

    def Bind(self, tag, f):
        if tag == 'parameter': self.handle = f if not f is None else print
        if tag == 'commit': self.on_ok = f
        if tag == 'cancel': self.on_cancel = f

    def show(self):
        if self.modal: 
            status =  self.ShowModal() == 5100
            self.Destroy()
            return status
        else: self.Show()

    def __del__( self ):
        print('panel config deleted!')

def get_para(para, view, title='Parameter', parent=None):
    pd = ParaDialog(parent, title)
    pd.init_view(view, para)
    pd.pack()
    rst = pd.ShowModal()
    pd.Destroy()
    return rst == 5100

if __name__ == '__main__':
    para = {'name':'yxdragon', 'age':10, 'h':1.72, 'w':70, 'sport':True, 'sys':'Mac', 'lan':['C/C++', 'Python'], 'c':(255,0,0)} 

    view = [('lab', 'lab', 'This is a questionnaire'),
            (str, 'name', 'name', 'please'), 
            (int, 'age', (0,150), 0, 'age', 'years old'),
            (float, 'h', (0.3, 2.5), 2, 'height', 'm'),
            ('slide', 'w', (1, 150), 0, 'weight','kg'),
            (bool, 'sport', 'do you like sport'),
            (list, 'sys', ['Windows','Mac','Linux'], str, 'favourite', 'system'),
            ('chos', 'lan', ['C/C++','Java','Python'], 'lanuage you like(multi)'),
            ('color', 'c', 'which', 'you like')]

    app = tk.Tk()
    pd = ParaDialog(None, 'Test')
    pd.init_view(view, para, preview=True, modal=False)
    
    print(para)
    app.mainloop()



import numpy as np
import moderngl
import numpy as np
from math import sin, cos, tan, pi

class Surface:
	def __init__(self, vts, ids, ns, cs=(0,0,1), **key):
		self.vts = vts
		self.ids = ids
		self.ns = ns
		self.cs = cs
		self.box = np.vstack(
			(vts.min(axis=0), 
			 vts.max(axis=0)))
		self.mode = 'mesh'
		self.alpha = 1
		self.visible = True
		self.width = 1
		self.update = False
		self.color = (0,0,0)
		self.set_style(**key)

	def set_style(self, **key):
		if 'mode' in key: self.mode = key['mode']
		if 'alpha' in key: self.alpha = key['alpha']
		if 'visible' in key: self.visible = key['visible']
		if 'color' in key:
			self.cs = key['color']
			self.update = True	

class MarkText(Surface):
	def __init__(self, vts, ids, os, h, color):
		self.vts = vts
		self.ids = ids
		self.cs = color
		self.os = os
		self.h = h
		self.box = None
		self.alpha = 1
		self.visible = True
		self.width = 1
		self.mode = 'grid'
		self.update = False
		self.color = (0,0,0)

class MeshSet:
	def __init__(self, name='meshset', objs=None):
		self.name = name
		self.objs = objs or {}

	@property
	def box(self):
		minb = [i.box[0] for i in self.objs.values() if not i.box is None]
		maxb = [i.box[1] for i in self.objs.values() if not i.box is None]
		minb, maxb = np.array(minb).min(axis=0), np.array(maxb).max(axis=0)
		return np.vstack((minb, maxb))

	@property
	def center(self): return self.box.mean(axis=0)

	@property
	def dial(self): return np.linalg.norm(self.box[1]-self.box[0])

	def add_surf(self, name, obj):
		self.objs[name] = obj

	def get_obj(self, key):
		if not key in self.objs: return None
		return self.objs[key]





import tkinter as tk
from tkinter import ttk
from tkinter import messagebox,filedialog
import os
from scitk.widgets.tkwidgetext.closablenotebook import ClosableNotebook
class TextPad(ttk.Frame):
    def __init__(self, parent, cont='', title='no name'):
        super().__init__(parent)
        #self.SetBackgroundColour( wx.SystemSettings.GetColour( wx.SYS_COLOUR_3DLIGHT ) )
        self.title = title
        self.savePath=''
        self.parent=parent
        
        self.text=tk.Text(self,undo=True)
        self.text.insert('current',cont)
        self.text.pack()
        self.notebook=None
        self.text.bind('<Button-3>',self.OnRClick)
        
        menus = [
                ## File 
                ('File(&F)',[('Open', self.OnOpen),
                                   ('Save', self.OnSave),
                                  ('Save as', self.OnSaveAs)
                                   ]),
                ## Edit 
                ('Edit(&E)', [ ('Undo', self.OnUndo),
                             ('Redo', self.OnRedo),
                             ('Cut', self.OnCut),
                             ('Copy', self.OnCopy),
                             ('Paste', self.OnPaste),
                             ('All', self.OnSelectAll)
                             ]),               
                ## Help 
                ('Help(&H)', [('About', self.OnAbout)])
        ]
        
        ### Bind menus with the corresponding events 
        self.popupMenu=tk.Menu()
        for menuTup in menus:
            parent=menuTup[0]
            for subcmd in menuTup[1]:
                self.popupMenu.add_command(label=subcmd[0], command=subcmd[1])
            self.popupMenu.add_separator()

    def OnOpen(self):
        filename= filedialog.askopenfilename(title='Select file to open',
                                            filetypes=[('Markdown', '*.md'), ('All Files', '*')]) 
        
        if os.path.isfile(filename):
            with open(filename) as f:
                self.text.delete('0.0',tk.END)
                self.text.insert('0.0',f.read())
            self.savePath=filename
            self.notebook.on_save_path_changed()


    def OnSaveAs(self):
        path=filedialog.asksaveasfilename(title='Choose where to save', defaultextension=".md") 
        if path:
           
            self.savePath=path
            self.notebook.on_save_path_changed()
            with open(path,'w') as f:
                f.write(self.text.get(1.0,tk.END))
                f.close()


    def OnSave(self):
        print(self.savePath,'save')
        if not os.path.isfile(self.savePath):
            self.OnSaveAs()
        else:
            with open(self.savePath,'w' ) as f:
                f.write(self.text.get(1.0,tk.END))

    def OnAbout(self,event):
        pass
        #wx.MessageBox('Text Log Window!','ImagePy',wx.OK)

    def OnRClick(self,event):
        self.popupMenu.tk_popup(event.x_root, event.y_root)

    def OnUndo(self): self.text.edit_undo()

    def OnRedo(self): self.text.edit_redo()

    def OnCut(self):self.text.event_generate("<<Cut>>")

    def OnCopy(self): self.text.event_generate("<<Copy>>")

    def OnPaste(self): self.text.event_generate("<<Paste>>")

    def OnSelectAll(self): self.text.tag_add(tk.SEL, "1.0", tk.END)
        
    def append(self, cont:str):
        self.text.insert("insert",cont)

class TextFrame(ttk.Frame):    
    def __init__(self, parent, title='no name', cont=''):
        super().__init__(parent)
        self.title = title
        self.textpad = TextPad(self, cont, title)
        self.append = self.textpad.append
        self.pack()
        
        ### Create menus (name:event) k-v pairs 
        menus = [
                ## File 
                ('File(&F)',[('Open', self.textpad.OnOpen),
                             ('Save', self.textpad.OnSave),
                             ('Save as', self.textpad.OnSaveAs),
                             ('-'),
                             ('Exit', self.OnClose)
                             ]),
                ## Edit 
                ('Edit(&E)', [ ('Undo', self.textpad.OnUndo),
                             ('Redo', self.textpad.OnRedo),
                             ('-'),
                             ('Cut', self.textpad.OnCut),
                             ('Copy', self.textpad.OnCopy),
                             ('Paste', self.textpad.OnPaste),
                             ('-'),
                             ('All', self.textpad.OnSelectAll)
                             ]),               
                ## Help 
                ('Help(&H)', [('About', self.textpad.OnAbout)])
        ]
        
        ### Bind menus with the corresponding events 
        self.popupMenu=tk.Menu()
        for menuTup in menus:
            parent=menuTup[0]
            for subcmd in menuTup[1]:
                self.popupMenu.add_command(label=subcmd[0], command=subcmd[1])
            self.popupMenu.add_separator()
##        for menu in menus:
##            m = wx.Menu()
##            for item in menu[1]:
##                if item[0]=='-':
##                    m.AppendSeparator()
##                else:
##                    i = m.Append(-1, item[0])
##                    self.Bind(wx.EVT_MENU,item[1], i)
##            self.menuBar.Append(m,menu[0])
##        self.SetMenuBar(self.menuBar) 
##        self.Bind(wx.EVT_CLOSE, self.OnClosing)

    def OnClose(self,event):
        self.Destroy()
        
    def OnClosing(self, event):
        event.Skip()
        
class TextNoteBook(ClosableNotebook):
    def __init__(self, parent):
        super().__init__(parent)
        self.bind("<<SavePathChanged>>",self.on_save_path_changed,add='+')
        
##        self.Bind( wx.lib.agw.aui.EVT_AUINOTEBOOK_PAGE_CHANGED, self.on_valid) 
##        self.Bind( wx.lib.agw.aui.EVT_AUINOTEBOOK_PAGE_CLOSE, self.on_close)
##        self.Bind( wx.EVT_IDLE, self.on_idle)
##        self.SetArtProvider(aui.AuiSimpleTabArt())
        
    def on_idle(self, event):
        for i in range(self.GetPageCount()):
            title = self.GetPage(i).title
            if self.GetPageText(i) != title:
                self.SetPageText(i, title)

    def textpad(self, i=None):
        self.get_tab(i)
        
    def set_background(self, img):
        self.GetAuiManager().SetArtProvider(ImgArtProvider(img))

    def add_notepad(self, textpad=None):
        if textpad==None:
            textpad=TextPad(self)
            self.add_tab(textpad,textpad.title)
        else:
            self.add_tab(textpad,textpad.title)
        print(textpad)
        textpad.notebook=self
        return textpad

    def set_title(self, panel, title):
        self.SetPageText(self.GetPageIndex(panel), title)

    def on_valid(self, event): pass

    def on_close(self, event): pass
    
    def on_save_path_changed(self):
        for index in range(len(self.tabs())):
            tabText=self.tabList[index].savePath
            self.tab(index,text=tabText)

class TextNoteFrame(ttk.Frame):
    def __init__(self,parent, title='TextPadBookFrame'):
        super().__init__ (parent)
        
        self.notebook = TextNoteBook(parent)
        self.notebook.pack(expand=True,fill=tk.BOTH) 
        self.textpad = self.notebook.textpad
        self.add_notepad=self.notebook.add_notepad
        
        ### Create menus (name:event) k-v pairs 
        menus = [
                ## File 
                ('File(&F)',[('Open', lambda e: self.textpad().OnOpen(e)),
                             ('Save', lambda e: self.textpad().OnSave(e)),
                             ('Save as', lambda e: self.textpad().OnSaveAs(e)),
                             ('-'),
                             ('Exit', self.OnClose)
                             ]),
                ## Edit 
                ('Edit(&E)', [ ('Undo', lambda e: self.textpad().OnUndo(e)),
                             ('Redo', lambda e: self.textpad().OnRedo(e)),
                             ('-'),
                             ('Cut', lambda e: self.textpad().OnCut(e)),
                             ('Copy', lambda e: self.textpad().OnCopy(e)),
                             ('Paste', lambda e: self.textpad().OnPaste(e)),
                             ('-'),
                             ('All', lambda e: self.textpad().OnSelectAll(e))
                             ]),               
                ## Help 
                ('Help(&H)', [('About', lambda e: self.textpad().OnAbout(e))])
        ]
        
        ### Bind menus with the corresponding events 
        self.menuBar=tk.Menu(self)
        for menu in menus:
            self.menuBar.add_command(label=menu[0],command=print)
        parent.config(menu=self.menuBar)
        return
        if i==1:
            for item in menu[1]:
                if item[0]=='-':
                    m.AppendSeparator()
                else:
                    i = m.Append(-1, item[0])
                    self.Bind(wx.EVT_MENU,item[1], i)
            self.menuBar.Append(m,menu[0])
        self.SetMenuBar(self.menuBar) 
        self.Bind(wx.EVT_CLOSE, self.OnClosing)

    def OnClose(self,event):
        self.Destroy()
        
    def OnClosing(self, event):
        event.Skip()
    
if __name__ == '__main__':
    app = tk.Tk()

    npbf = TextNoteFrame(app)
    note1 = npbf.add_notepad()
    note1.append('abc')
    note2 = npbf.add_notepad()
    note2.append('def')
    npbf.pack()
    app.mainloop()
    




from time import time
import numpy as np
from math import pi

lib = {'0':([(0,0.5,0.5,0,0)],[(1,1,0,0,1)],0.5),
	'1':([(0.25,0.25)], [(0,1)], 0.5),
	'2':([(0,0.5,0.5,0,0,0.5)], [(1,1,0.5,0.5,0,0)], 0.5),
	'3':([(0,0.5,0.5,0),(0,0.5)],[(1,1,0,0),(0.5,0.5)], 0.5),
	'4':([(0,0,0.5),(0.5,0.5)],[(1,0.5,0.5),(1,0)],0.5),
	'5':([(0.5,0,0,0.5,0.5,0)], [(1,1,0.5,0.5,0,0)], 0.5),
	'6':([(0.5,0,0,0.5,0.5,0,0)], [(1,1,0.5,0.5,0,0,0.5)], 0.5),
	'7':([(0,0.5,0.5)], [(1,1,0)], 0.5),
	'8':([(0.5,0.5,0,0,0.5,0.5,0,0)], [(0.5,1,1,0.5,0.5,0,0,0.5)], 0.5),
	'9':([(0.5,0.5,0,0,0.5,0.5,0)], [(0.5,1,1,0.5,0.5,0,0)], 0.5),
	'I':([(0,0.5),(0.25,0.25),(0,0.5)],[(1,1),(1,0),(0,0)],0.5),
	'D':([(0,0.25,0.4,0.5,0.5,0.4,0.25,0),(0.1,0.1)],[(1,1,0.9,0.75,0.25,0.1,0,0),(0,1)],0.5),
	':':([(0.2,0.3),(0.2,0.3)],[(0.75,0.75),(0.25,0.25)],0.5)}

def count_ns(vts, fs):
	dv1 = vts[fs[:,1]] - vts[fs[:,2]]
	dv2 = vts[fs[:,1]] - vts[fs[:,0]]
	ns = np.cross(dv1, dv2)
	ass = np.linalg.norm(ns, axis=1)
	ns /= np.linalg.norm(ns, axis=1).reshape((-1,1))
	buf = np.zeros_like(vts)
	for i in (0,1,2): np.add.at(buf, fs[:,i], ns)
	buf /= np.linalg.norm(buf, axis=1).reshape((-1,1))
	return buf

def build_twringidx(n, offset=0):
	idx = np.array([[0,1,n+1],[n+1,n+2,1]])
	idx = idx[np.arange(n*2)%2].T + np.arange(n*2)//2
	return (idx.T+offset).astype(np.uint32)

def build_pringidx(p, n, offset=0):
	ridx = np.array([[0,0,1]]*n, dtype=np.uint32)
	ridx += np.arange(n, dtype=np.uint32).reshape((-1,1))+offset
	ridx[:,0] = p
	return ridx

def build_grididx(r, c):
	idx = np.arange(r*c, dtype=np.uint32)
	rs, cs = idx//c, idx%c
	idx1 = idx[(rs<r-1)*(cs<c-1)].reshape((-1,1))
	did = np.array([[0, 1, 1+c, 0, 1+c, c]], dtype=np.uint32)
	return rs, cs, (idx1 + did).reshape((-1,3))

def build_surf2d(img, ds=1, sigma=0, k=0.2):
	from skimage.filters import sobel_h, sobel_v
	from scipy.ndimage import gaussian_filter
	#start = time()
	img = img[::-ds, ::ds]
	img = gaussian_filter(img, sigma)
	r, c = img.shape
	rs, cs, fs = build_grididx(r, c)
	vs = img[rs, cs]

	vts = np.array([cs*ds, rs*ds, vs*k], dtype=np.float32).T
	cs = (np.ones((3, r*c))*(vs/255)).astype(np.float32).T
	
	dx, dy = sobel_h(img), sobel_v(img)
	cx, cy = np.zeros((r*c, 3)), np.zeros((r*c, 3))
	cx[:,0], cx[:,2] = 1, dx.ravel()
	cy[:,1], cy[:,2] = 1, dy.ravel()
	ns = np.cross(cx, cy)
	ns = (ns.T/np.linalg.norm(ns, axis=1)).astype(np.float32).T
	return vts, fs, ns, cs

def build_surf3d(imgs, ds, level, step=1, c=(1,0,0)):
	from skimage.measure import marching_cubes_lewiner
	vts, fs, ns, cs =  marching_cubes_lewiner(imgs[::ds,::ds,::ds], level, step_size=step)
	vts *= ds
	cs = (np.ones((len(vts), 3))*c).astype(np.float32)
	return vts, fs, ns, cs

def build_ball(o, r, c=(1,0,0)):
	ay, ax = np.mgrid[-pi/2:pi/2:9j, 0:pi*2:17j]
	zs = np.sin(ay.ravel())
	xs = np.cos(ax.ravel()) * np.cos(ay.ravel())
	ys = np.sin(ax.ravel()) * np.cos(ay.ravel())
	ns = np.vstack((xs, ys, zs)).astype(np.float32).T
	vts = (ns * r + o).astype(np.float32)
	fs = build_grididx(9, 17)[2]
	cs = (np.ones((len(vts), 3))*c).astype(np.float32)
	#print(ns2)
	return vts, fs, ns, cs

def build_mesh(xs, ys, zs, c=(1,0,0)):
	rs, cs, fs = build_grididx(xs.shape[0], xs.shape[1])
	vts = np.array([xs[rs, cs], ys[rs, cs], zs[rs,cs]]).astype(np.float32).T
	ns = count_ns(vts, fs)
	cs = (np.ones((len(vts), 3))*c).astype(np.float32)
	return vts, fs, ns, cs

def build_balls(os, rs, cs=(1,0,0)):
	if isinstance(cs, tuple):
		cs = [cs] * len(os)
	vtss, fss, nss, css = [], [], [], []
	for o,r,c in zip(os, rs, cs):
		vv, ff, nn, cc = build_ball(o, r, c)
		fss.append(ff+len(vtss)*len(vv))
		vtss.append(vv)
		nss.append(nn)
		css.append(cc)
	return np.vstack(vtss), np.vstack(fss), np.vstack(nss), np.vstack(css)

# 0 1 1  2 2 3  3 4 4  5 5 6
def build_line(xs, ys, zs, c):
	vts = np.array([xs, ys, zs], dtype=np.float32).T
	n = (len(xs)-1)*2
	rem = (6 - n % 6)%6
	fs = np.arange(0.1,(n+rem)//2,0.5).round().astype(np.uint32)
	if rem>0: fs[-rem:] = len(xs)-1 
	ns = np.ones((len(vts), 3), dtype=np.float32)
	cs = (np.ones((len(vts), 3))*c).astype(np.float32)
	return vts, fs.reshape((-1,3)), ns, cs

def build_lines(xs, ys, zs, cs):
	if not isinstance(cs, list):
		cs = [cs] * len(xs)
	vtss, fss, nss, css = [], [], [], []
	s = 0
	for x, y, z, c in zip(xs, ys, zs, cs):
		vv, ff, nn, cc = build_line(x, y, z, c)
		fss.append(ff+s)
		s += len(vv)
		vtss.append(vv)
		nss.append(nn)
		css.append(cc)
	return np.vstack(vtss), np.vstack(fss), np.vstack(nss), np.vstack(css)

def build_arrow(v1, v2, rs, re, ts, te, c):
	v = (v2-v1)/np.linalg.norm(v2-v1)
	ss, ee = v1 + v*rs*ts, v2 - v*re*te
	vx = np.cross(v, np.random.rand(3))
	vx /= np.linalg.norm(vx)
	vy = np.cross(vx, v)
	angs = np.linspace(0, np.pi*2, 17)
	vas = np.array([np.cos(angs), np.sin(angs)])
	vxy = np.dot(vas.T, np.array([vx, vy]))
	vts = np.vstack((v1, ss + rs * vxy, ee + re * vxy, v2))
	fs1 = build_pringidx(0, 16, 1)
	fs = build_twringidx(16, 1)
	fs2 = build_pringidx(35, 16, 18)
	face = np.vstack((fs1, fs, fs2))
	ns = np.vstack((-v, vxy, vxy, v)).astype(np.float32)
	cs = (np.ones((len(vts), 3))*c).astype(np.float32)
	return vts.astype(np.float32), face, ns, cs

def build_arrows(v1s, v2s, rss, res, tss, tes, cs):
	if not isinstance(cs, list): cs = [cs] * len(v1s)
	if not isinstance(tss, list): tss = [tss] * len(v1s)
	if not isinstance(tes, list): tes = [tes] * len(v1s)
	if not isinstance(rss, list): rss = [rss] * len(v1s)
	if not isinstance(res, list): res = [res] * len(v1s)
	vtss, fss, nss, css = [], [], [], []
	s = 0
	for v1, v2, rs, re, ts, te, c in zip(v1s, v2s, rss, res, tss, tes, cs):
		if np.linalg.norm(v1-v2) < 0.1: continue
		vv, ff, nn, cc = build_arrow(v1, v2, rs, re, ts, te, c)
		fss.append(ff+s)
		s += len(vv)
		vtss.append(vv)
		nss.append(nn)
		css.append(cc)
	print(np.vstack(vtss).shape, np.vstack(fss).shape, np.vstack(nss).shape, np.vstack(css).shape)
	return np.vstack(vtss), np.vstack(fss), np.vstack(nss), np.vstack(css)

def build_mark(cont, pos, dz, h, color):
	vts, fss = [], []
	s, sw = 0, 0
	for i in cont:
		xs, ys, w = lib[i]
		vv, ff, nn, cc = build_lines(xs, ys, ys, (0,0,0))
		fss.append(ff+s)
		vts.append(vv+[sw,0,0])
		vts[-1][:,2] = dz
		s += len(vv)
		sw += w+0.3
	sw -= 0.3
	vts = (np.vstack(vts)-[sw/2.0, 0.5, 0])
	return vts, np.vstack(fss), pos, h, color

def build_marks(conts, poss, dz, h, color):
	if not hasattr(dz, '__len__'):
		dz = [dz] * len(conts)
	vtss, fss, pps = [], [], []
	s = 0
	for cont, pos, z in zip(conts, poss, dz):
		vv, ff, pp, hh, cc = build_mark(cont, pos, z, h, color)
		fss.append(ff+s)
		s += len(vv)
		vtss.append(vv)
		pps.append((np.ones((len(vv),3))*pp).astype(np.float32))

	return np.vstack(vtss), np.vstack(fss), np.vstack(pps), h, color

def build_cube(p1, p2, color=(1,1,1)):
	(x1,y1,z1),(x2,y2,z2) = p1, p2
	xs = (x1,x2,x2,x1,x1,x1,x1,x1,x1,x2,x2,x1,x2,x2,x2,x2)
	ys = (y1,y1,y1,y1,y1,y2,y2,y1,y2,y2,y2,y2,y2,y1,y1,y2)
	zs = (z1,z1,z2,z2,z1,z1,z2,z2,z2,z2,z1,z1,z1,z1,z2,z2)
	return build_line(xs, ys, zs, color)

def build_img_cube(imgs, ds=1):
	imgs = imgs[::ds,::ds,::ds]
	(h, r, c), total = imgs.shape[:3], 0
	print(h, r, c)
	vtss, fss, nss, css = [], [], [], []
	shp = [(h,r,c,h*r), (h,c,r,h*c), (r,c,h,r*c)]
	nn = [[(0,0,-1),(0,0,1)], [(0,1,0),(0,-1,0)], [(1,0,0),(-1,0,0)]]
	for i in (0,1,2):
		rs, cs, fs12 = build_grididx(*shp[i][:2])
		idx1, idx2 = [rs*ds, cs*ds], [rs*ds, cs*ds]
		rcs1, rcs2 = [rs, cs], [rs, cs]
		rcs1.insert(2-i, 0); rcs2.insert(2-i, -1)
		vs1, vs2 = imgs[tuple(rcs1)]/255, imgs[tuple(rcs2)]/255
		idx1.insert(2-i, rs*0); idx2.insert(2-i, cs*0+shp[i][2]*ds-1)
		vtss.append(np.array(idx1, dtype=np.float32).T)
		vtss.append(np.array(idx2, dtype=np.float32).T)
		css.append((np.ones((1, 3))*vs1.reshape((len(vs1),-1))).astype(np.float32))
		css.append((np.ones((1, 3))*vs2.reshape((len(vs1),-1))).astype(np.float32))
		nss.append((np.ones((shp[i][3],1))*[nn[i][0]]).astype(np.float32))
		nss.append((np.ones((shp[i][3],1))*[nn[i][1]]).astype(np.float32))
		fss.extend([fs12+total, fs12+(total+shp[i][0]*shp[i][1])])
		total += shp[i][3] * 2
	return np.vstack(vtss), np.vstack(fss), np.vstack(nss), np.vstack(css)
	
def build_img_box(imgs, color=(1,1,1)):
	return build_cube((-1,-1,-1), imgs.shape[:3], color)

cmp = {'rainbow':[(127, 0, 255), (43, 126, 246), (42, 220, 220), (128, 254, 179), (212, 220, 127), (255, 126, 65), (255, 0, 0)],
	'jet':[(0, 0, 127), (0, 40, 255), (0, 212, 255), (124, 255, 121), (255, 229, 0), (255, 70, 0), (127, 0, 0)],
	'ocean':[(0, 127, 0), (0, 64, 42), (0, 0, 85), (0, 64, 128), (0, 127, 170), (129, 192, 213), (255, 255, 255)],
	'earth':[(0, 0, 0), (27, 77, 122), (54, 135, 111), (93, 160, 75), (169, 179, 91), (206, 171, 132), (253, 250, 250)]}

def linear_color(cs):
	if isinstance(cs, str): cs=cmp[cs]
	cmap = np.zeros((256, 3), dtype=np.uint8)
	idx = np.linspace(0, 256, len(cs)).astype(np.uint16)
	for i in range(1, len(cs)):
		c1, c2 = cs[i-1], cs[i]
		rs, gs, bs = [np.linspace(c1[j], c2[j], idx[i]-idx[i-1]) for j in (0,1,2)]
		cmap[idx[i-1]:idx[i]] = np.array((rs, gs, bs)).T
	return cmap

def auto_lookup(vs, cmap):
	vs = vs - vs.min()
	vs = vs/vs.max()
	vs = (vs*255).astype(np.uint8)
	return cmap[vs]

if __name__ == '__main__':
	from matplotlib import cm
	cmap = linear_color('earth')
	import matplotlib.pyplot as plt
	img = np.zeros((30,256), dtype=np.uint8)
	img[:] = np.arange(256)
	img = cmap[img]
	plt.imshow(img)
	plt.show()


from tkinter import ttk
import tkinter as tk
from PIL import ImageTk,Image

def make_logo(logo_path:str,logo_size):
    '''
    The lib tkinter only supports .gif format as button logo. This function is used to convert any type of logo file to show on tkinter widgets.
    tkinter 只支持.gif格式的图像作为按钮的图标。这个函数可以将任意格式的图像文件转换为可以显示在控件上的图像。
    '''
    img_origin = Image.open(logo_path)  # 打开图片
    if not type(logo_size)==tuple: 
        raise Exception("Invalid logo size.logo_size should be a tuple such as (24,24)")
    img_origin=  img_origin.resize(logo_size, Image.ANTIALIAS)     # 缩放到所需大小
    img = ImageTk.PhotoImage(img_origin)  # 用PIL模块的PhotoImage打开

    return img
    
class Toolbar(ttk.Frame):
    '''
    Toolbar(ttk.Frame) 
    input params in __init__ method:  parent
    accessible properties(可以访问的属性):
    --logo_size  :the size of logo,is a tuple(w,h),default is (24,24)
    '''
    def __init__(self,parent):
        super().__init__(parent)
        self.logo_size = (24,24) # logo的尺寸。
        self.__image_set = set()# 对各个命令对应的按钮进行引用。

    def add_tool(self,logo:str,tool:callable):
        '''
        logo:str,图片路径
        tool:callable,点击此工具时调用的函数。
        '''
        img = make_logo(logo,self.logo_size)
        self.__image_set.add(img)
        b=ttk.Button(self,text='ddddd',image =img,command = tool)
        b.pack(side=tk.LEFT)


if __name__=="__main__":
    app = tk.Tk()
    tb = Toolbar(app)
    tb.pack(fill=tk.X,expand =True)
    tb.add_tool('/media/hzy/程序/novalide/forgitcommit/NovalIDE/noval/bmp_source/unittest_wizard.png',lambda:print('ddd'))
    text = tk.Text(app)
    text.pack(fill=tk.BOTH,expand=True)
    app.mainloop()
    


'''
    inspired by yxl
    code by hzyrc6011
    
    要查看示例，请直接运行此文件。
    To view the demo, please directly run this file.
    
'''
import tkinter as tk
from tkinter import ttk


class ScrolledCanvas(ttk.Frame):
    '''
    这是一个用来在可以滚动的区域中容纳其他控件的Frame容器。
    This is a Frame container used for containing other widgets in a scrollable area.
    
    注意：需要添加的控件不能直接放置(place/grid/pack)在此控件上，而应该放置在其scrollable_frame上。
    Attention:  Widgets need to be added cannot be directly put (place/grid/pack) on this frame ,
    but should be put on the scrollable_frame of this widget.
    '''
    def __init__(self, container,orient = 'v',  *args, **kwargs):
        super().__init__(container, *args, **kwargs)
        self.canvas = tk.Canvas(self)

       
        self.scrollable_frame = ttk.Frame(self.canvas)

        self.scrollable_frame.bind(
            "<Configure>",
            lambda e: self.canvas.configure(
                scrollregion= self.canvas.bbox("all")
            )
        )

        self.canvas.create_window((0, 0), window=self.scrollable_frame, anchor="nw")

        

       
        if orient.startswith('v'):
            self.canvas.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
            self.scrollbar = ttk.Scrollbar(self, orient="vertical", command=self.canvas.yview)
            self.canvas.configure(yscrollcommand=self.scrollbar.set)
            self.scrollbar.pack(side="right", fill="y")
            
        elif orient.startswith('h'):
            self.canvas.pack(side=tk.TOP, fill=tk.BOTH, expand=True)
            self.scrollbar = ttk.Scrollbar(self, orient="horizontal", command=self.canvas.xview)    
            self.canvas.configure(xscrollcommand=self.scrollbar.set)        
            self.scrollbar.pack(side=tk.BOTTOM, fill="x")
            
        else:
            raise('Unknown scrollbar orientation \'{0}\' (未知的滚动条位置设置：{0})'%(orient) )
        #canvas.configure(yscrollcommand=self.scrollbar.set)
if __name__ == "__main__":
    app=tk.Tk()
    sf=ScrolledCanvas(app,'h')
    sf.pack(expand=1,fill=tk.BOTH)
    for i in range(50):
        ttk.Button(sf.scrollable_frame, text="Sample scrolling label%d"%i).pack(side=tk.LEFT)
    app.mainloop()



import numpy as np

class Table():
    def __init__(self, df=None, name='Table'):
        self.name = name
        self.df = df
        self.rg = None
        self.props = None
        self.snap = None
        self.dirty = False
        if not df is None:
            self.data = df
        
    @property
    def title(self): return self.name

    @property
    def nbytes(self):
        return self.data.memory_usage().sum()

    @property
    def columns(self):return self.data.columns

    @property
    def index(self):return self.data.index

    @property
    def shape(self): return self.data.shape

    @property
    def data(self): return self.df
    
    def update(self): self.dirty = True

    @data.setter
    def data(self, df):
        self.df, self.props = df, None
        self.rowmsk, self.colmsk = [], []
        self.count_range()

    @property
    def style(self):
        props, data = self.props, self.data
        if props is None or set(props)!=set(data.columns):
            ps = [[3, (0,0,0), (0,0,255), 'Text'] for i in data.columns]
            self.props = dict(zip(data.columns, ps))
        return self.props

    def subtab(self, mode=True, num=False):
        rs, cs = len(self.rowmsk), len(self.colmsk)
        mskr = slice(None) if rs==0 else self.rowmsk
        mskc = slice(None) if cs==0 else self.colmsk
        if mode is None: mskr = mskc = slice(None)
        mskr, mskc = self.data.index[mskr], self.data.columns[mskc]
        if mode is False and rs>0: mskr = self.data.index.difference(mskr)
        if mode is False and cs>0: mskc = self.data.columns.difference(mskc)
        subtab = self.data.loc[mskr, mskc]
        if num: subtab = subtab.select_dtypes(include=[np.number])
        return subtab

    def snapshot(self, mode=True, num=False):
        self.snap = self.subtab(mode, num).copy()

    def set_style(self, col, **key):
        for i, name in enumerate(['accu', 'tc', 'lc', 'ln']):
            if name in key: self.style[col][i] = key[name]

    def count_range(self):
        rg = list(zip(self.data.min(), self.data.max()))
        self.rg = dict(zip(self.data.columns, rg))

    def select(self, rs=[], cs=[], byidx=False):
        if byidx: rs, cs = self.df.index[rs], self.df.columns[cs]
        self.rowmsk, self.colmsk = rs, cs

    @property
    def info(self):
        return '%sx%s %.2fM'%(*self.shape, self.nbytes/1024/1024)
        
if __name__ == '__main__':
    import numpy as np
    import pandas as pd
    df = pd.DataFrame(np.zeros((10,5)), columns=list('abcde'))
    table = Table(df)
    



from tkinter import ttk
import tkinter as tk
from scitk.widgets.tkwidgetext.choicebook import BaseChoiceBook
class ChoiceBook(ttk.Frame):
    def __init__(self, parent):
        super().__init__(parent)
        
    def add_wgts(self, name, wgts):
        book = BaseChoiceBook(self)
        for name, wgt in wgts:
            book.add(wgt(book),name)
        book.pack(expand = 1, fill=tk.BOTH )

    def load(self, data):
        for name, wgts in data[1]:
            self.add_wgts(name, wgts)

class TestFrame(ttk.Frame):
    def __init__(self,parent):
        super().__init__(parent)
        self.b=ttk.Button(self,text='哈哈哈哈')
        self.b.pack()
if __name__ == '__main__':
    app = tk.Tk()
    frame = ttk.Frame(None)
    book = ChoiceBook(frame)
    book.load(('widgets', [('panels', [('A', TestFrame), ('B', ttk.Frame)]),
        ('panels2', [('A', ttk.Frame), ('B', ttk.Frame)])]))
    book.pack(expand=True,fill=tk.BOTH)
    frame.pack(expand=True, fill=tk.BOTH)
    app.mainloop()



from .app import App
from .manager import Manager

class Source:
    managers = {}

    @classmethod
    def manager(cls, name):
        if not name in cls.managers: 
            cls.managers[name] = Manager()
        return cls.managers[name]


import tkinter as tk
import cv2
import PIL.Image, PIL.ImageTk
import time
import numpy as np
# Create a window
window = tk.Tk()
window.title("OpenCV and Tkinter")

# Load an image using OpenCV
cv_img2 = cv2.cvtColor(cv2.imread("/home/hzy/Desktop/深度截图_选择区域_20200613232201.png"), cv2.COLOR_BGR2RGB)
cv_img=np.column_stack((cv_img2,cv_img2))
print(cv_img.shape)
 # Get the image dimensions (OpenCV stores image data as NumPy ndarray)
height, width, no_channels = cv_img.shape

class Canvas(tk.Canvas):
    def __init__(self,master,width,height):
        super().__init__(master,width = width, height = height)
        self.images=[]
        self.image=None
        self.imgtmp=None
        self.imgpos=(0,0)
        self.relativePos=(0,0)
        self.bind('<B1-Motion>',self.showImage)
        self.bind('<ButtonRelease-1>',self.endDrag)
        self.bind('<ButtonPress-1>',self.startDrag)
        self.fps=0
        self.t1=self.t0=time.time()
        self.img=cv_img
    def startDrag(self,event):
        self.relativePos=(-event.x+self.imgpos[0],-event.y+self.imgpos[0])
    def endDrag(self,event):
        pass
    def showImage(self,event):
        global cv_img,photo
        self.t1=time.time()
        if self.t1-self.t0>=1:
            print(self.fps)
            self.fps=0
            self.t0=time.time()
        self.imgtmp = PIL.ImageTk.PhotoImage(image = PIL.Image.fromarray(cv_img))
        self.create_image(event.x+self.relativePos[0],event.y+ self.relativePos[1], image=self.imgtmp, anchor=tk.NW)
        self.fps+=1
        
# Create a canvas that can fit the above image
canvas = Canvas(window, width = width, height = height)
canvas.pack(side=tk.LEFT)
##c2= Canvas(window, width = width, height = height)
##c2.pack(side=tk.LEFT)
# Use PIL (Pillow) to convert the NumPy ndarray to a PhotoImage

 
 # Add a PhotoImage to the Canvas

window.mainloop()
